"""
This module contains some wrappers for using ASE to perform
various operations on BigDFT molecules.

https://wiki.fysik.dtu.dk/ase/
"""
from ase.calculators.calculator import Calculator


def bigdft_to_ase(sys):
    """
    Given a BigDFT system, this transforms that system into an ASE
    collection of atoms.

    Args:
      frag (BigDFT.Systems.System): the system to convert.

    Returns:
      (ase.Atoms): a collection of atoms for ASE.
    """
    from ase import Atom, Atoms
    from ase.units import Bohr

    atlist = []
    for frag in sys.values():
        for at in frag:
            pos = at.get_position("angstroem")
            atlist += [Atom(at.sym, (pos[0], pos[1], pos[2]))]

    cellargs = {}
    if sys.cell is not None and sys.cell.get_boundary_condition() != "free":
        cellargs["cell"] = []
        cellargs["pbc"] = []

        for i in range(3):
            if sys.cell[i, i] == float("inf"):
                cellargs["cell"].append([0, 0, 0])
                cellargs["pbc"].append(False)
            else:
                cellargs["cell"].append([x*Bohr for x in sys.cell[i, :]])
                cellargs["pbc"].append(True)

    return Atoms(atlist, **cellargs)


def ase_to_bigdft(mol):
    """
    Given an ASE collection of atoms, this transforms that collection into
    a BigDFT fragment.

    Args:
      mol (ase.Atoms): a collection of atoms used by ASE.

    Returns:
      (BigDFT.Fragments.Fragment): a BigDFT fragment.
    """
    from BigDFT.Fragments import Fragment
    from BigDFT.Atoms import Atom

    frag = Fragment()
    for at in mol:
        frag += [Atom({"sym": at.symbol, "r": at.position,
                       "units": "angstroem"})]

    return frag


def ase_potential_energy(sys, ase_calculator):
    """
    Given a ASE calculator, calculates the potential energy
    of a BigDFT System.

    Args:
        sys (BigDFT.Systems.System): a BigDFT system
        ase_calculator (ase.calculators.calculator.Calculator): ASE calculator

    Returns:
        float: the potential energy, in Hartree
    """
    from ase.units import Hartree
    asys = bigdft_to_ase(sys)
    asys.set_calculator(ase_calculator)
    return asys.get_potential_energy() / Hartree


class BigASECalculator(Calculator):
    """
    This calculator can be used by ASE to run BigDFT calculations.

    Args:
      parameters (BigDFT.Inputfiles.Inputfile): the input parameters to use.
      calc (BigDFT.Calculators.SystemCalculator): the calculator to use.
      directory (str): the directory to run the calculation in.
      label (str): a label for this particular calculation.
      atoms (ase.Atoms): an ase collection of atoms to compute.
    """
    def __init__(self, parameters, calc, restart=None,
                 ignore_bad_restart_file=False, directory=".", label=None,
                 atoms=None, **kwargs):

        # Call the Super Method
        Calculator.__init__(self, restart, ignore_bad_restart_file,
                            label, atoms, **kwargs)

        # Internal data.
        self.log = None
        self.code = calc
        self.parameters = parameters
        self.rundir = directory

    def calculation_required(self, atoms, quantities):
        """
        https://wiki.fysik.dtu.dk/ase/_modules/ase/calculators/interface.html
        """
        from copy import deepcopy

        if atoms != self.atoms:
            self.atoms = deepcopy(atoms)
            return True
        elif self.log is None:
            return True
        else:
            return False

    def get_forces(self, atoms):
        """
        https://wiki.fysik.dtu.dk/ase/_modules/ase/calculators/interface.html
        """
        from numpy import array
        from ase.units import Bohr, Hartree

        if self.calculation_required(atoms, quantities=["forces"]):
            self._run_calculation()
        sys = self._create_system()
        sys.set_atom_forces(self.log)

        forces = []
        for at in sys["FRAG:0"]:
            forces.append(at.get_force())

        self.results["forces"] = array(forces) * (Hartree / Bohr)
        return self.results["forces"]

    def get_potential_energy(self, atoms=None, force_consistent=False):
        """
        https://wiki.fysik.dtu.dk/ase/_modules/ase/calculators/interface.html
        """
        from ase.units import Hartree
        if self.calculation_required(atoms, quantities=["energy"]):
            self._run_calculation()
        self.results["energy"] = self.log.energy * Hartree
        return self.results["energy"]

    def _create_system(self):
        """
        Create a BigDFT system from the atoms of this calculator.
        """
        from BigDFT.Systems import System
        from BigDFT.UnitCells import UnitCell
        from ase.units import Bohr

        sys = System()
        sys["FRAG:0"] = ase_to_bigdft(self.atoms)

        if self.atoms.cell is not None:
            if any(self.atoms.get_pbc()):
                if not self.atoms.cell.orthorhombic:
                    raise ValueError("Only orthorhombic cells are supported.")
                vec = [float(self.atoms.cell[i, i]) if self.atoms.get_pbc()[i]
                       else float("inf") for i in range(3)]
                sys.cell = UnitCell(vec, units="angstroem")

        return sys

    def _get_posinp(self):
        """
        Get the posinp associated with the atoms of this calculator.
        """
        return self._create_system().get_posinp()

    def _run_calculation(self):
        self.log = self.code.run(input=self.parameters,
                                 posinp=self._get_posinp(),
                                 name=self.label, run_dir=self.rundir)


def _example():
    """Example of using ASE interoperability"""
    from BigDFT.IO import XYZReader
    from BigDFT.Inputfiles import Inputfile
    from BigDFT.Calculators import SystemCalculator
    from BigDFT.Fragments import Fragment
    from BigDFT.Systems import System
    from ase.calculators.lj import LennardJones
    from BigDFT.UnitCells import UnitCell
    from ase.units import Hartree
    from ase.optimize import BFGS
    from os.path import join
    from copy import deepcopy

    # Create a system.
    sys = System()
    reader = XYZReader("Ar")
    sys["FRA:1"] = Fragment(xyzfile=reader)
    sys["FRA:2"] = deepcopy(sys["FRA:1"])
    sys["FRA:2"].translate([-2, 0, 0])

    # Skip straight to the potential energy.
    print(ase_potential_energy(sys, LennardJones()))

    # Advanced used.
    asys = bigdft_to_ase(sys)
    asys.set_calculator(LennardJones())
    dyn = BFGS(asys)
    dyn.run(fmax=0.05)
    print(asys.get_potential_energy() / Hartree)

    # Unit cells
    sys.cell = UnitCell([5, 5, 5], units="bohr")
    print(ase_potential_energy(sys, LennardJones()))
    sys.cell = UnitCell([5, float("inf"), 5], units="bohr")
    print(ase_potential_energy(sys, LennardJones()))
    sys.cell = UnitCell([float("inf"), float("inf"), 5], units="bohr")
    print(ase_potential_energy(sys, LennardJones()))

    # We can also use BigDFT with ase.
    inp = Inputfile()
    inp.set_xc("PBE")
    inp.set_hgrid(0.4)
    code = SystemCalculator(verbose=False)
    sys.cell = UnitCell()
    print(ase_potential_energy(sys, BigASECalculator(inp, code,
                                                     directory="work",
                                                     label="ase-free")))
    sys.cell = UnitCell([5, 5, 5], units="bohr")
    print(ase_potential_energy(sys, BigASECalculator(inp, code,
                                                     directory="work",
                                                     label="ase-periodic")))
    sys.cell = UnitCell([5, float("inf"), 5], units="bohr")
    print(ase_potential_energy(sys, BigASECalculator(inp, code,
                                                     directory="work",
                                                     label="ase-surface")))
    sys.cell = UnitCell([float("inf"), float("inf"), 5], units="bohr")
    print(ase_potential_energy(sys, BigASECalculator(inp, code,
                                                     directory="work",
                                                     label="ase-wire")))


if __name__ == "__main__":
    _example()



"""
This module has the routines and data structures necessary to allow one to
generate visualizations of a atomic systems. We also include helper routines
for generating colors.

"""


class InlineVisualizer():
    """
    This class allows for a quick viewing of BigDFT systems using the
    py3Dmol package.

    https://pypi.org/project/py3Dmol/

    Attributes:
      xsize (int): the width of the picture in pixels.
      ysize (int): the height of the picture in pixels.
      nrow (int): if present, the number of rows for displaying a grid of
          structures.
      ncol (int): if present, the number of columns for displaying a grid of
          structures.
    """

    def __init__(self, xsize, ysize, nrow=1, ncol=1):
        from py3Dmol import view
        # set linked to False so that each structure can be rotated
        # independently
        self.xyzview = view(width=ncol*xsize, height=nrow*ysize,
                            viewergrid=(nrow, ncol), linked=False)

    def display_system(self, *syslist, **kwargs):
        """
        Display an animation of a sequence of systems. The colordict can be
        used to color each fragment. When only one system is passed it
        will remain still.

        Args:
          syslist (BigDFT.Systems.System): the systems to visualize.
          colordict (dict): a dictionary from fragment ids to hex colors,
             can also be a list of dicts (one for each system) if using a grid.
          field_vals(list): values of the field to decide colors of the keys
          cartoon (bool): set to True to use the cartoon view. This only works
            if atom names and residues are properly defined.
          gridlist (list): if present, defines the row and column indices for
            visualizing multiple systems on a grid.
          show (bool): you can explicitly defer showing.
        """
        from BigDFT.IO import write_pdb
        # py2 workaround
        from sys import version_info
        if version_info[0] < 3:
            from io import BytesIO as StringIO
        else:
            try:
                from io import StringIO
            except ImportError:
                from StringIO import StringIO

        colordict = kwargs.get('colordict')
        field_vals = kwargs.get('field_vals')
        cartoon = kwargs.get('cartoon', False)
        gridlist = kwargs.get('gridlist', None)
        stick_radius = kwargs.get('stick_radius', 0.0)
        stick_color = kwargs.get('stick_color', 'black')
        show = kwargs.get('show', True)

        # Set the default colors
        if colordict is None:
            keyset = []
            for system in syslist:
                keyset += list(system.keys())
            keyset = list(set(keyset))
            colordict = get_colordict(keyset, field_vals=field_vals)

        # Draw each system
        models = ""
        for s, system in enumerate(syslist):
            if type(colordict) is list:
                this_colordict = colordict[s]
            else:
                this_colordict = colordict

            if type(stick_color) is list:
                this_stick_color = stick_color[s]
            else:
                this_stick_color = stick_color

            model = ""
            sval = StringIO()
            write_pdb(system, sval)
            model += "MODEL " + str(s) + "\n"
            model += sval.getvalue()
            model += "ENDMDL\n"

            # in this case we have only one (combined) system to display
            if gridlist is None:
                models += model
            # in this case we just want to display this system at the moment
            else:
                models = model

            # If displaying on a grid, remove all existing models
            if gridlist is not None:
                gx = gridlist[s][0]
                gy = gridlist[s][1]
                viewer = (gx, gy)
                # print(viewer)
                self.xyzview.removeAllModels(viewer=viewer)
            else:
                viewer = (0, 0)

            # in the case of a grid we need to display at each iteration
            # otherwise we only display for the final iteration
            if (gridlist is not None) or (gridlist is None and
                                          s == len(syslist) - 1):
                i = 0
                for fragid, frag in syslist[0].items():
                    if fragid in this_colordict:
                        color = this_colordict[fragid]
                    else:
                        color = 'black'

                    if not cartoon:
                        if stick_radius > 0.0:
                            self.xyzview.addModelsAsFrames(models, "pdb",
                                                           {"keepH": "true"},
                                                           viewer=viewer)
                            self.xyzview.setStyle({'model': -1},
                                                  {"stick":
                                                  {'radius': stick_radius,
                                                   'color': this_stick_color}},
                                                  viewer=viewer)
                        else:
                            self.xyzview.addModelsAsFrames(models, "pdb",
                                                           {"keepH": "true"},
                                                           viewer=viewer)
                            self.xyzview.setStyle({'model': -1},
                                                  {"line": {'color': 'black'}},
                                                  viewer=viewer)

                    for at in frag:
                        if cartoon:
                            self.xyzview.addModelsAsFrames(models,
                                                           viewer=viewer)
                            self.xyzview.setStyle({'model': -1,
                                                   'serial': i + 1},
                                                  {"cartoon":
                                                  {'color': color}},
                                                  viewer=viewer)
                        else:

                            self.xyzview.addModelsAsFrames(models, "pdb",
                                                           {"keepH": "true"},
                                                           viewer=viewer)
                            self.xyzview.setStyle({'model': -1, 'serial': i+1},
                                                  {"sphere": {'scale': 0.2,
                                                              'color': color}},
                                                  viewer=viewer)
                        i += 1

            if gridlist is not None:
                self.xyzview.render()
        self.display_cell(syslist[0].cell)

        # Finish Drawing
        if len(syslist) > 1 and gridlist is None:
            self.xyzview.animate({'loop': "forward", 'interval': 1000})
        self.xyzview.zoomTo()

        if show:
            self.xyzview.show()

    def display_cell(self, cell):
        if cell is None:
            return
        self.xyzview.addUnitCell({"model": -1},
                                 {"box": {"color": "black"},
                                 "alabel": "", "blabel": "", "clabel": ""})


class VSimVisualizer():
    def __init__(self, filename, xsize=600, ysize=600):
        from gi.repository import v_sim, GLib, Gtk
        self.win = v_sim.UiRenderingWindow.new(xsize, ysize, True, True)
        self.loop = GLib.MainLoop.new(None, False)
        self.win.connect_object('destroy', GLib.MainLoop.quit, self.loop)
        self.main = Gtk.Window.new(Gtk.WindowType.TOPLEVEL)
        self.main.add(self.win)
        self.main.show_all()
        self._set_file(filename)

    def _set_file(self, filename):
        from gi.repository import v_sim
        self.data = v_sim.DataAtomic.new(filename, None)
        self.data.load(0, None)
        self.win.getGlScene().setData(self.data)
        # v_sim.basic_parseConfigFiles()

    def show(self):
        # from gi.repository import GLib
        self.loop.run()

    def colorize_by_fragments(self):
        from gi.repository import v_sim
        self._scene_colorizer(v_sim, self.data, self.win.getGlScene())

    def _scene_colorizer(self, v_sim, data, scene):
        nodes = scene.getNodes()
        frag = data.getNodeProperties("Fragment")
        c = v_sim.DataColorizerFragment.new()
        c.setNodeModel(frag)
        nodes.pushColorizer(c)
        c.setActive(True)

    def colorizer_script(self, filename):
        import inspect
        towrite = ['scene = ' +
                   'v_sim.UiMainClass.getDefaultRendering().getGlScene()',
                   'data = scene.getData()']
        for line in inspect.getsource(self._scene_colorizer).split('\n')[1:]:
            towrite.append(line.lstrip(' '))
        f = open(filename, 'w')
        for line in towrite:
            f.write(line + '\n')
        f.close()


class VMDGenerator():
    """
    This class contains the routines you would use for visualization of
    a system using the VMD program.

    Attributes:
      representation (str): the vmd representation to draw with.
        https://www.ks.uiuc.edu/Research/vmd/allversions/repimages/#representations
      color (int): the default color to draw with.
    """

    def __init__(self, representation="CPK", color=16):
        self.representation = representation
        self.color = color

    def visualize_fragments(self, system, scriptfile, geomfile,
                            fragcolors=None):
        """
        This generates a script for visualizing the fragmentation of a
        system using VMD.

        Args:
          system (BigDFT.Systems.System): the system to visualize.
          scriptfile (str): the name of the file to write the vmd script
            to (usually has extension .tcl)
          geomfile (str): the filename for where to write an xyz file
            of the system.
          fragcolors (dict): optionally, a dictionary from fragment ids to
            fragment colors. Colors are integers between 0 and 32.
        """
        from BigDFT.Fragments import Fragment
        from BigDFT.IO import XYZWriter

        # To create the XYZ file, we first make one big fragment.
        geomorder = Fragment()
        for fragid, frag in system.items():
            geomorder += frag

        # Then write it to file.
        with XYZWriter(geomfile, len(geomorder)) as ofile:
            for at in geomorder:
                ofile.write(at)

        # Get the matching so we can write the correct atom indices.
        matching = system.compute_matching(geomorder)

        # If fragcolors is not specified, we will generate it ourselves.
        if fragcolors is None:
            fragcolors = {}
            for i, s in enumerate(system):
                # 16 is black, which we will reserve.
                if i % 32 == 16:
                    c = str(32)
                else:
                    c = str(i % 32)
                fragcolors[s] = c

        # The header of the script file draws the whole system in black.
        outstr = self._get_default_header(geomfile)

        # This part colors individual fragments.
        modid = 1
        for fragid, frag in system.items():
            if fragid not in fragcolors:
                continue
            outstr += "mol addrep 0\n"
            outstr += """mol modselect """ + str(modid) + """ 0 index """
            outstr += " ".join([str(x) for x in matching[fragid]])
            outstr += "\n"
            outstr += """mol modcolor """
            outstr += str(modid) + """ 0 ColorID """ + \
                str(fragcolors[fragid]) + """\n"""
            modid += 1

        # Finally, write to file.
        with open(scriptfile, "w") as ofile:
            ofile.write(outstr)

    def _get_default_header(self, geomfile):
        outstr = """mol default style """+self.representation+"""\n"""
        outstr += """mol new """ + geomfile + "\n"
        outstr += """mol modcolor 0 0 ColorID """+str(self.color)+"""\n"""

        return outstr


def get_distinct_colors(keys, name="tab20", fuzz=True):
    """
    This generates a dictionary of distinct colors based on a matplotlib
    colormap.

    Args:
        keys (list): a list of keys.
        name (str): the name of the matplotlib colormap to use.
        fuzz (bool): some color maps (included tab20) only have a distinct
            set of colors. The fuzz option adds increased randomness to make
            up for this.

    Returns:
        (dict): a dictionary mapping matplotlib keys to RGB colors.
    """
    from matplotlib import pyplot as plt
    from numpy import linspace
    from random import sample, uniform

    def fuzz(x):
        fuzzed = (x[i] + uniform(-0.1, 0.1) for i in range(3))
        return [max(0.0, min(y, 1.0)) for y in fuzzed]

    # Map using a color map
    cmap = plt.get_cmap(name)
    xvals = linspace(0, 1, len(keys))
    vals = {x: list(cmap(xvals[i])) for i, x in
            enumerate(sample(keys, len(keys)))}

    # Add some random fuzz
    if fuzz:
        vals = {x: fuzz(y) for x, y in vals.items()}

    return vals


def get_colordict(keys, field_vals=None, colorcode='html'):
    """
    Build a dictionary of colors for each of the keys.
    If the field_dict is provided, order the colors of the keys in terms
    of the sorting of the filed values

    Args:
        keys (list): keys of the color dictionary
        field_vals(list) : values of the field to decide the colors of the keys
        colorcode (str): the string of the colorcode.
             Can be 'html' or 'rgb' if field_dict is absent,
             otherwise it represent the colormap of matplotlib.
             Default is 'seismic' for diverging values
             (field_dict has negative data), otherwise 'Reds'
    Returns:
        dict: the dictionary of the keys, and the corresponding colors
    """
    import numpy as np
    from matplotlib.pyplot import get_cmap
    cvals = _find_colours(len(keys))
    if field_vals is None:
        colorkey = 'indices' if colorcode == 'rgb' else colorcode
        cvals = cvals[colorkey]
    else:
        compressed_values = np.array(
            [val for val in field_vals if not np.isnan(val)])
        center_to_zero = any(compressed_values < 0.0)
        mx = np.max(compressed_values)
        mn = np.min(compressed_values)
        if center_to_zero:
            shift = min(mn, -mx)
            top = max(mx, -mn)
        else:
            shift = mn
            top = mx
        compressed_values -= shift
        if top > shift:
            compressed_values /= (top-shift)
        if colorcode == 'html':
            colorkey = 'seismic' if center_to_zero else 'Reds'
        else:
            colorkey = colorcode
        cmap = get_cmap(colorkey)
        cvals = []
        icv = 0
        for v in field_vals:
            if np.isnan(v):
                cvals.append('None')
            else:
                y = compressed_values[icv]
                cvals.append(_rgb_to_html(
                    tuple(map(int, np.array(cmap(y)[0:3])*255))))
                icv += 1
        # cvals = [_rgb_to_html(tuple(map(int, np.array(cmap(y)[0:3])*255)))
        #          for y in compressed_values]

    return {x: y for x, y in zip(keys, cvals)}


def _find_colours(n):
    """
    A helper routine for finding a list of N colors in varous formats.

    Args:
      n (int): the number of colors to generate.
           Cycles colours after 1408 values

    Returns:
      (dict): a dictionary describing the colors. The key `html` contains
      the hex values values, and `indices` contains a list of RGB values.
    """
    colours = {}
    colours['html'] = list()
    colours['indices'] = list()
    ind3d = [0] * 3
    for i in range(n):
        ind1d = 128+i*int(1408/n)
        if (ind1d < 256):
            ind3d[0] = ind1d
            ind3d[1] = 0
            ind3d[2] = 0
        elif (ind1d < 512):
            ind3d[0] = 255
            ind3d[1] = ind1d-256
            ind3d[2] = 0
        elif (ind1d < 768):
            ind3d[0] = 768-ind1d-1
            ind3d[1] = 255
            ind3d[2] = 0
        elif (ind1d < 1024):
            ind3d[0] = 0
            ind3d[1] = 255
            ind3d[2] = ind1d-768
        elif (ind1d < 1280):
            ind3d[0] = 0
            ind3d[1] = 1280-ind1d-1
            ind3d[2] = 255
        elif (ind1d <= 1536):
            ind3d[0] = ind1d-1280
            ind3d[1] = 0
            ind3d[2] = 255

        colours['indices'].append(ind3d[:])
        html = _rgb_to_html(ind3d)

        colours['html'].append(html)

    return colours


def _rgb_to_html(rgb):
    html = [0] * 6
    htmls = [0] * 6
    html[0] = int(rgb[0]/16)
    html[1] = (rgb[0] % 16)
    html[2] = int(rgb[1]/16)
    html[3] = (rgb[1] % 16)
    html[4] = int(rgb[2]/16)
    html[5] = (rgb[2] % 16)
    for j in range(6):
        if (html[j] == 10):
            htmls[j] = 'A'
        elif (html[j] == 11):
            htmls[j] = 'B'
        elif (html[j] == 12):
            htmls[j] = 'C'
        elif (html[j] == 13):
            htmls[j] = 'D'
        elif (html[j] == 14):
            htmls[j] = 'E'
        elif (html[j] == 15):
            htmls[j] = 'F'
        else:
            htmls[j] = html[j]
        html_string = '#'+str(htmls[0])+str(htmls[1]) + \
            str(htmls[2])+str(htmls[3])+str(htmls[4])+str(htmls[5])
    return html_string


def _example():
    """Visualization Example"""
    from BigDFT.Systems import System
    from BigDFT.Fragments import Fragment
    from BigDFT.IO import XYZReader

    # Read in a system.
    sys = System()
    with XYZReader("SiO") as ifile:
        for i, at in enumerate(ifile):
            sys["FRA:"+str(i)] = Fragment([at])

    # Display the system.
    viz = InlineVisualizer(400, 300)
    viz.display_system(sys)

    # Change the colors
    colordict = get_distinct_colors(list(sys))
    viz.display_system(sys, colordict=colordict)


if __name__ == "__main__":
    _example()



"""
This module contains the System level class of PyBigDFT. Systems are named
collections of fragments, and represent a complete system for simulation.
"""
from futile.Utils import write as safe_print

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping


class System(MutableMapping):
    """
    A system is defined as a named collection of fragments. You can manipulate
    a system as if it were a standard python dictionary, however it also has
    helper routines for performing operations on the full system.
    """

    def __init__(self, *args, **kwargs):
        from BigDFT.UnitCells import UnitCell
        self.store = dict()
        self.update(dict(*args, **kwargs))
        self.conmat = None
        self.cell = UnitCell()

    def dict(self):
        """
        Convert to a dictionary.
        """
        return self.store

    def __getitem__(self, key):
        return self.store[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self.store[self.__keytransform__(key)] = value

    def __delitem__(self, key):
        del self.store[self.__keytransform__(key)]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __keytransform__(self, key):
        return key

    def __eq__(self, other):
        """
        Compare two systems. They are equal if all the fragments of
            the systems are identical.

        other (System): the fragment to compare with
        """
        ok = True
        for fr1 in self:
            ok = ok and self[fr1] == other[fr1]
            if not ok:
                break
        return ok

    @property
    def centroid(self):
        """
        Center of mass of the system
        """
        from numpy import mean
        return mean([frag.centroid for frag in self.values()], axis=0)

    @property
    def central_fragment(self):
        """
        Returns the fragment whose center of mass is closest to the centroid

        Returns:
          (str): the name of the fragment.
          (Fragment): the fragment object
        """
        import numpy as np
        CMs = [frag.centroid for frag in self.values()]
        idx = np.argmin([np.dot(dd, dd.T) for dd in (CMs - self.centroid)])
        return list(self.keys())[idx], list(self.values())[idx]

    def get_external_potential(self, units="bohr", charge_offset=False):
        """
        Transform the system information into a dictionary ready to be
        put as an external potential.

        Args:
          units (str): the units of the external potential.
          charge_offset (bool): by default the external potential ignores the
            counter charge from the protons. Setting this to true adds the
            positive charge to the potential.
        """
        ret_dict = {"units": units}
        ret_dict["values"] = []
        for frag in self.values():
            ret_dict["values"].extend(
                frag.get_external_potential(units, charge_offset))
        # ret_dict["global monopole"] = sum(
        #     x["q0"][0] for x in ret_dict["values"])
        return ret_dict

    def get_k_nearest_fragments(self, target, k, cutoff=None):
        """
        Given a fragment id in a system, this computes the nearest fragment.

        Args:
          target (str): the fragment to find the nearest neighbor of.
          k (int): the number of fragments to look for.
          cutoff (float): will only return fragments with a certain cutoff.

        Returns:
          (list): the ids of the nearest fragments.
        """
        from scipy.spatial import KDTree

        # Setup the KD Tree for distance lookup
        poslist = []
        frag_lookup = []
        for fragid, frag in self.items():
            if fragid == target:
                continue
            for at in frag:
                poslist.append(at.get_position())
                frag_lookup.append(fragid)
        tree = KDTree(poslist)

        # Find the nearest fragments with a query of the tree.
        targetpost = [x.get_position() for x in self[target]]
        if cutoff is not None:
            ndist, nearest = tree.query(targetpost, k=k,
                                        distance_upper_bound=cutoff)
        else:
            ndist, nearest = tree.query(targetpost, k=k)

        # We now have the nearest atom to each atom in this fragment.
        # Next we combine this information and extract the closest
        # fragments.

        if k == 1:
            ndist = [ndist]
            nearest = [nearest]

        distdict = {}
        for i in range(0, len(nearest)):
            for idx, dist in zip(nearest[i], ndist[i]):
                try:
                    fragidx = frag_lookup[idx]
                except IndexError:
                    # kdtree returns invalid indices if it can't find enough
                    # points.
                    continue
                if fragidx not in distdict:
                    distdict[fragidx] = dist
                elif distdict[fragidx] > dist:
                    distdict[fragidx] = dist

        # Extract the k smallest values.
        minlist = []
        for i in range(0, k):
            if len(distdict) == 0:
                break
            key = min(distdict, key=distdict.get)
            minlist.append(key)
            del distdict[key]

        return minlist

    def get_nearest_fragment(self, target):
        """
        Given a fragment id in a system, this computes the nearest fragment.

        Args:
          target (str): the fragment to find the nearest neighbor of.

        Returns:
          (str): the id of the nearest fragment.
        """
        return self.get_k_nearest_fragments(target, k=1)[0]

    def get_net_force(self):
        """
        Returns the net force on a system in Ha/Bohr.

        Returns:
          (list): Three values which describe the net force.
        """
        from numpy import array
        ret_val = array([0.0, 0.0, 0.0])
        for frag in self.values():
            ret_val += array(frag.get_net_force())
        return [float(x) for x in ret_val]

    def get_posinp(self, units='angstroem'):
        """
        Provide the dictionary which has to be passed to the ``posinp`` value
        of the :meth:`run` method of  the
        :class:`~BigDFT.Calculators.SystemCalculator` class instance.

        Args:
           units (str): The units of the file. May be "angstroem" or "bohr".
        """
        pos = []
        for fragid, frag in self.items():
            for at in frag:
                atdict = {at.sym: at.get_position(units)}
                atdict["frag"] = list(GetFragTuple(fragid))
                if frag.frozen:
                    atdict["Frozen"] = frag.frozen
                pos.append(atdict)

        result = {}
        result["units"] = units
        result["positions"] = pos
        result["cell"] = self.cell.get_posinp(units)

        return result

    def serialize(self, order=None, units='bohr'):
        """
        Transform the system in a list that can be employed for
        the construction of dataframes or pandas series.
        Args:
            order (list): list of fragments to serialize in order
            units (str): the units for the positions
        Returns:
            list: the serialized system as well as a lookup dictionary
                that contains the order of the fragment atoms in the series
        """
        if order is None:
            order = self
        positions = []
        lookup = {}
        iat = 0
        for key in order:
            frag = self[key]
            lookup[key] = [iat+i for i in range(len(frag))]
            iat += len(frag)
            positions += frag.serialize(name=key, units=units)
        return positions

    @property
    def df(self):
        if not hasattr(self, '_df'):
            self._df = self.to_dataframe()
        return self._df

    def to_dataframe(self, **kwargs):
        """
        Convert the system into a dataframe, from the
        `py:meth:~System.serialize` method.

        Args:
            **kwargs: arguments to be passed to
                `py:meth:System.serialize` method
        """
        from pandas import DataFrame as DF
        df = DF(self.serialize(**kwargs))
        validate_dataframe_representation(self, df)
        return df

    def dataframe_slicing(self, df=None):
        """
        Define a dictionaries of two-element lists indicating the slicing
        (start and end points) of the corresponding fragment in the order
        provided by the dataframe

        Args:
            df (Dataframe): associated to the system

        Returns:
            dict:  dictionary of the slicings of the fragments in the dataframe
        """
        if df is None:
            df = self.df
        sl = fragment_slicing_from_system_dataframe(df)
        check_slicing(self, sl)
        return sl

    @property
    def PointParticles(self):
        """
        Transform the system into a `py:class:~PointParticles.PointParticles`
        object
        """
        from BigDFT.PointParticles import PointParticles as PP
        if not hasattr(self, '_PP'):
            self._PP = PP(**point_particle_objects(self.df))
        return self._PP

    @property
    def electrostatic_interactions(self):
        """Dictionary of the Electrostatic interactions between fragments.
        """
        return self.PointParticles.Eel_dict(self.dataframe_slicing())

    @property
    def q0(self):
        """
        Provides the global monopole of the system given as a sum of the
        monopoles of the atoms.
        """
        if len(self) == 0:
            return None
        return [sum(filter(None, [frag.q0[0] for frag in self.values()]))]

    @property
    def qcharge(self):
        """
        The total qcharge of a system.
        """
        return sum([frag.qcharge for frag in self.values()])

    def rename_fragments(self, fragment_mapping=None):
        """
        This procedure automatically names the fragments in a system.

        Args:
           fragment_mapping (dict) : Dictionary containing list
                 of fragments that are additionally added to each
                 of the original system's fragments.

        Returns:
          BigDFT.Systems.System: the same system, with the automatic naming
          scheme.
        """
        rnsys = System()
        for i, fragid in enumerate(self):
            if fragment_mapping is None:
                tuplek = "FRAG:"+str(i)
            else:
                additional_list = fragment_mapping[fragid]
                if not isinstance(additional_list, list):
                    tuplek = additional_list
                elif len(additional_list) == 0:
                    tuplek = fragid
                else:
                    # tuplek = (fragid, ) + tuple((f for f in additional_list))
                    tuplek = fragid + '+'+'+'.join(
                             [f for f in additional_list])
            rnsys[tuplek] = self[fragid]
        return rnsys

    def set_atom_multipoles(self, logfile, correct_charge=True):
        """
        After a run is completed, we have a set of multipoles defined on
        each atom. This routine will set those values on to each atom
        in the system.

        Args:
          logfile (Logfiles.Logfile): logfile with the multipole values.
          correct_charge (bool): currently there is an inconsistency in
            terms of gross charge, and this corrects it.
        """
        mp = logfile.electrostatic_multipoles
        for pole in mp["values"]:
            pole["units"] = mp["units"]
        lookup = self.compute_matching(mp["values"])

        # Assign
        for fragid, frag in self.items():
            for i, at in enumerate(frag):
                idx = lookup[fragid][i]
                if idx >= 0:
                    at.set_multipole(mp["values"][idx], correct_charge)

    def set_atom_forces(self, logfile):
        """
        After a run is completed, we have the forces on each atom in the
        logfile. This routine will set those values to each atom in this sytem.

        Args:
          logfile (Logfiles.Logfile): logfile with the forces.
        """
        from BigDFT.Fragments import Fragment
        posinp = logfile.log.get('posinp')
        if posinp is not None and not isinstance(posinp, str):
            atlist = Fragment(posinp=posinp)
        else:
            atlist = Fragment(astruct=logfile.astruct)
        lookup = self.compute_matching(atlist)

        # Assign forces
        try:
            forces = logfile.forces
        except AttributeError:
            forces = logfile.astruct["forces"]
        for fragid, frag in self.items():
            for i, at in enumerate(frag):
                idx = lookup[fragid][i]
                if idx >= 0:
                    at.set_force(list(forces[idx].values())[0])

    def compute_matching(self, atlist, shift=None, check_matching=True):
        """
        Frequently we are passed a list of atom like objects from which we
        need to extract data and assign it to a system. However, a system
        can potentially store those atoms in any order, and may not have
        the same set of atoms. This helper routine creates a mapping between
        this list view, to the dictionary view of the system class.

        Args:
          atlist (list): a list of atom like objects.
          shift (list): if the positions in atlist are shifted by some constant
            vector you can specify that here.
          check_matching (bool): if set to True, this will raise an error
            if we can't match all of the atoms in the system.

        Returns:
          (dict): a mapping from a system to indices in the atom list. If
            an atom is not in the list, an index value of -1 is assigned.
        """
        from BigDFT.Atoms import Atom
        from numpy import array
        from scipy.spatial import KDTree

        # Convert everything to pure positions to avoid overhead.
        poslist = [array(Atom(x).get_position("bohr")) for x in atlist]
        if shift is not None:
            poslist = [x - array(shift) for x in poslist]
        tree = KDTree(poslist)

        # Seach for the mapping values
        mapping = {}
        for fragid, frag in self.items():
            mapping[fragid] = []
            for at in frag:
                atpos = array(at.get_position("bohr", cell=self.cell))
                ndist, nearest = tree.query(atpos)
                if check_matching and ndist > 0.01:
                    raise ValueError("Unable to match atom" + str(dict(at)))
                mapping[fragid].append(nearest)

        return mapping

    def display(self, colordict=None, field_vals=None, cartoon=False):
        """
        Display the system using the inline visualizer of Py3DMol
        """
        from BigDFT.Visualization import InlineVisualizer
        viz = InlineVisualizer(400, 300)
        viz.display_system(self, colordict=colordict, field_vals=field_vals,
                           cartoon=cartoon)
        return viz

    def set_electrons_from_log(self, log):
        """
        BigDFT uses pseudopotentials, so this will extract from a logfile
        the number of actual electrons modelled for each atom in the system.

        Args:
           log (Logfiles.Logfile) : A BigDFT run logfile.
        """
        electrons = {}
        for key in log.log:
            if "psppar" not in key:
                continue
            electrons[key.split(".")[1]] = log.log[key]['No. of Electrons']

        for frag in self.values():
            for at in frag:
                at.nel = electrons[at.sym]

    def set_logfile_info(self, log):
        """
        Include the information of the logfile in the fragment quantities.

        Args:
           log (Logfiles.Logfile) : A BigDFT run logfile.
        """
        # provide the atomic information on the system
        if hasattr(log, 'electrostatic_multipoles'):
            self.set_atom_multipoles(log)
        if hasattr(log, 'forces'):
            self.set_atom_forces(log)
        self.set_electrons_from_log(log)

    def ase_potential_energy(self, ase_calculator):
        """
        Given a ASE calculator, calculates the potential energy
        of the system.

        Args:
            ase_calculator (ase.calculators.calculator.Calculator): ASE
              calculator.

        Returns:
            float: the potential energy, in Hartree
        """
        from Fragments import Fragment
        from ASEInterop import ase_potential_energy as asepot
        bigfrag = Fragment(system=self)
        return asepot(bigfrag, ase_calculator)

    def reform_superunits(self, mapping):
        """
        Creates a new system from the provided mapping

        Args:
            mapping(dict): dictionary of the form {newfrag: [frag1,frag2]}
                defining the mapping between the old fragments and the new

        Returns:
            BigDFT.Systems.System: a new system from the remapping
        """
        newsys = System()

        for newfrag, fraglist in mapping.items():
            newsys[newfrag] = sum(self[f] for f in fraglist)

        return newsys

    def update_positions_from_dict(self, posinp):
        """
        Update the atomic positions of a system from a posinp dictionary.

        This method only works if the order of atoms match.

        Args:
            posinp (dict): a posinp dictionary.
        """
        from BigDFT.Atoms import Atom
        units = posinp.get("units", "angstroem")
        i = 0
        for frag in self.values():
            for at in frag:
                at2 = Atom(posinp["positions"][i], units=units)
                at.set_position(at2.get_position())
                i += 1


class Lattice():
    """
    Defines the fundamental objects to deal with periodic systems
    """

    def __init__(self, vectors):
        self.vectors = vectors

    def grid(self, origin=[0.0, 0.0, 0.0], extremes=None, radius=None):
        """
        Produces a set of translation vectors from a given origin
        """
        import numpy as np
        transl = []
        g = [[], [], []]  # the grid of discrete translations
        if extremes is not None:
            # print extremes
            for i, d in enumerate(extremes):
                for k in range(d[0], d[1] + 1):
                    g[i].append(k)
            # print g
            for i in g[0]:
                arri = np.array(self.vectors[0]) * i
                for j in g[1]:
                    arrj = np.array(self.vectors[1]) * j + arri
                    for k in g[2]:
                        arrk = np.array(self.vectors[2]) * k + arrj
                        vect = np.array(origin) + arrk
                        app = True
                        if radius is not None:
                            app = np.linalg.norm(arrk) < radius
                        if app:
                            transl.append(vect)
        return transl


def fragment_slicing_from_system_dataframe(df):
    """
    Define the slicing tuple needed to identify the fragment blocks
    into a system dataframe

    Args:
        df (Dataframe): the system dataframe

    Returns:
        dict: dictionary of the slicing obtained in form of  [start,end] list
    """
    current = df['frag'][0]
    slicing = {current: [0]}
    for i, f in enumerate(df['frag']):
        if f not in slicing:
            slicing[current].append(i)
            slicing[f] = [i]
            current = f
    slicing[current].append(i+1)
    return slicing


def check_slicing(system, slicing):
    """
    Assert the validity of a system's slicing
    """
    for frag, ss in slicing.items():
        assert len(system[frag]) == (ss[1]-ss[0])


def point_particle_objects(df):
    """
    Return the dictionary of the point particle quantities from a
    Systems' dataframe
    """
    if 'x_coord' in df:
        X = df[['x_coord', 'y_coord', 'z_coord']].to_numpy()
    else:
        X = df[['r_0', 'r_1', 'r_2']].to_numpy()
    P = df[['q1_2', 'q1_0', 'q1_1']].to_numpy()
    if 'qel_0' not in df:
        from BigDFT.Atoms import _nzion_default_psp as nzion
        from numpy import array
        df['zion'] = array([nzion[s] for s in df['sym']])
        # we assume net monopole definition
        df['qel_0'] = df['q0_0'] - df['nel']
    Q = df[['qel_0']].to_numpy()
    Z = df[['nel']].to_numpy()
    return {'X': X, 'P': P, 'Q': Q, 'Z': Z}


def _get_coordinate_keys(df):
    """Keys associated to atom coordinate in system's dataframe.
    """
    if 'x_coord' in df:
        return ['x_coord', 'y_coord', 'z_coord']
    else:
        return ['r_0', 'r_1', 'r_2']


def validate_dataframe_representation(sys, df):
    """
    Control if the units of the positions in the dataframe are in Bohr.
    Update the coordinate values if this is not so.

    Args:
        sys (BigDFT.System): System which provides the reference positions
        df (pandas.DataFrame): system's dataframe
    """
    import numpy as np
    from BigDFT.Atoms import AU_to_A
    from pandas import DataFrame
    assert all(df['units'] == 'bohr'), 'Dataframe units should be in Bohr'
    sl = sys.dataframe_slicing(df)
    keys = _get_coordinate_keys(df)
    all_coords = df[keys].to_numpy()
    for frag in sl:
        ist, ien = sl[frag]
        dfarr = all_coords[ist:ien]
        for at, coords in zip(sys[frag].atoms, dfarr):
            rxyz = np.array(at.get_position('bohr'))
            delta = rxyz - coords
            if np.linalg.norm(delta) > 1.e-3:  # this means angstroem
                dfarr /= AU_to_A
                delta = rxyz - coords
            assert np.linalg.norm(delta) <= 1.e-3, 'Atoms are not equivalent'
    df.update(DataFrame(all_coords, columns=keys))


class FragmentView():
    """
    The representation of a system in terms of fragments and
    groups of superunits.

    Args:
        purities (dict): dictionary of the purities of the system
        bond_order (dict): double dictionary of the bond orders of
            the superunits
        charges (dict): dictionary of the number of the electrons of
            the superunits
    """
    def __init__(self, purities, bond_orders, charges):
        self.purities = purities
        self.bond_orders = bond_orders
        self.charges = charges

    def __deepcopy__(self, memo):
        """
        Here we manually override deepcopy for performance reasons.
        """
        new_bo = {}
        new_pv = {}
        new_charges = {}
        for f1 in self.bond_orders:
            new_bo[f1] = {}
            for f2 in self.bond_orders[f1]:
                new_bo[f1][f2] = self.bond_orders[f1][f2]
            new_pv[f1] = self.purities[f1]
            new_charges[f1] = self.charges[f1]

        copy = FragmentView(new_pv, new_bo, new_charges)
        memo[id(self)] = copy
        return copy

    def refragment(self, mapping):
        newp, newbo = update_purity_and_bo(mapping, self.purities,
                                           self.bond_orders, self.charges)
        newchg = {}
        for refrag, remap in mapping.items():
            newchg[refrag] = sum(self.charges[f] for f in remap)
        return FragmentView(newp, newbo, newchg)

    def remove_fragment(self, fragid):
        """
        Remove a particular fragment from this view.

        Args:
          fragid (str): the id of the fragment to remove.
        """
        self.purities.pop(fragid)
        self.charges.pop(fragid)
        self.bond_orders.pop(fragid)

        for f in self.bond_orders:
            self.bond_orders[f].pop(fragid)


def select_from_view(view, targets):
    """
    Identify the fragments of the view that contain
    at least one of the targets

    Args:
        view (dict): if present, identifies the fragments that contain the
            relevant units
        targets (list): list of the fragments to search in the view
    Returns:
        list: fragments to select
    """
    reselect = []
    for frag in targets:
        for key, val in view.items():
            if frag in val and key not in reselect:
                reselect.append(key)
                break
    return reselect


def system_from_dict_positions(posinp, units='angstroem'):
    """
    Build a system from a set of positions from a dictionary whose yaml
    serialisation is compliant with the BigDFT yaml position format

    Args:
       posinp (list): list of the atomic specifications
    Returns:
       BigDFT.Systems.System: an instance of the system class.
          The employed fragment specification is specified in the file.
    """
    from BigDFT.Atoms import Atom
    from BigDFT.Fragments import Fragment
    sys = System()
    for iat, at in enumerate(posinp):
        frag = GetFragId(at, iat)
        if frag not in sys:
            sys[frag] = Fragment()
        sys[frag].append(Atom(at, units=units))
    return sys


def system_from_log(log, fragmentation=None):
    """
    This function returns a :class:`~BigDFT.Fragment.System` class out of a
    logfile. If the logfile contains information about fragmentation and atomic
    multipoles, then the system is created accordingly.
    Otherwise, the fragmentation scheme is determined by the fragmentation
    variable.

    Args:
       log (Logfile): the logfile of the QM run. In general must have been done
           with Linear Scaling formalism.
       fragmentation (str): the scheme to be used for the fragmentation in the
           case if not provided internally by the logfile.
           The possible values are ``atomic`` and ``full``, in which case the
           system as as many fragments as the number of atoms, or only one
           fragment, respectively.
    Returns:
        (BigDFT.Systems.System): The instance of the class containing
        fragments.
    """
    from BigDFT.Fragments import Fragment
    name = log.log.get('run_name', 'FULL') + ':0'

    full_system = System()
    posinp = log.log.get('posinp')
    if posinp is not None and not isinstance(posinp, str):
        full_system[name] = Fragment(posinp=posinp)
    else:
        full_system[name] = Fragment(astruct=log.astruct)

    full_system.set_logfile_info(log)

    # now we may defragment the system according to the provided scheme
    if fragmentation == 'full':
        return full_system
    elif fragmentation == 'atomic' or 'posinp' not in log.log:
        atomic_system = System()
        for iat, at in enumerate(full_system[name]):
            atomic_system['ATOM:' + str(iat)] = Fragment([at])
        return atomic_system
    else:
        posinp = log.log['posinp']
        frag_dict = {}
        for iat, tupl in enumerate(zip(posinp['positions'],
                                       full_system[name])):
            at, obj = tupl
            fragid = at.get('frag', 'ATOM:' + str(iat))
            if isinstance(fragid, list):
                fragid = ':'.join(map(str, fragid))
            if fragid not in frag_dict:
                frag_dict[fragid] = [obj]
            else:
                frag_dict[fragid].append(obj)
        frag_system = System()
        for fragid in frag_dict:
            frag_system[fragid] = Fragment(frag_dict[fragid])
        return frag_system


def plot_fragment_information(axs, datadict, colordict=None):
    """
    Often times we want to plot measures related to the different fragments
    in a system. For this routine, you can pass a dictionary mapping
    fragment ids to some kind of value. This routine takes care of the
    formatting of the axis to easily read the different fragment names.

    Args:
      axs (matplotlib.Axes): an axes object to plot on.
      datadict (dict): a dictionary from fragment ids to some kind of data
        value.
      colordict (dict): optionally, a dictionary from fragment ids to a
        color value.
    """
    # Sort by fragment id
    slabels = sorted(datadict.keys(),
                     key=lambda x: int(GetFragTuple(x)[1]))
    svalues = [datadict[x] for x in slabels]

    # Label the axis by fragments
    axs.set_xlabel("Fragment", fontsize=12)
    axs.set_xticks(range(len(datadict.keys())))
    axs.set_xticklabels(slabels, rotation=90)

    # Plot the actual values
    axs.plot(svalues, 'x', markersize=12, color='k')

    # Plot the colored values.
    if colordict:
        for i, key in enumerate(slabels):
            if key not in colordict:
                continue
            axs.plot(i, svalues[i], 'x', markersize=12, color=colordict[key])


def GetFragId(atdict, iat):
    """
    Obtain the fragment identifications from the atom description

    Args:
      atdict(dict): dictionary of the atom
      iat (int): position of the atom in the list

    Returns:
      str: fragment_id
    """
    fragid = atdict.get('frag', 'ATOM:' + str(iat))

    if all(':' in elem for elem in fragid):
        fragid = '+'.join(map(str, fragid))
    elif len(fragid) == 2 and ':' in str(fragid[1]):
        fragid = '-'.join(map(str, fragid))
    elif isinstance(fragid, list):
        fragid = ':'.join(map(str, fragid))
    return fragid


def GetFragTuple(fragid):
    """
    Fragment ids should have the form: "NAME:NUMBER" or "NAME-NUMBER". This
    splits the fragment into the name and number value.

    Args:
      fragid (str): the fragment id string.

    Return:
      (tuple): fragment name, fragment number
    """
    markers = [":", "-", "+"]
    if all(x not in fragid for x in markers):
        raise ValueError("Invalid format for fragment ID")
    if '+' in fragid:
        return tuple(fragid.split("+"))
    # elif '-' in fragid:
    #     return tuple(fragid.split("-"))
    elif isinstance(fragid, str):
        return tuple(fragid.split(":"))
    else:
        return tuple(fragid)


def copy_bonding_information(sys1, sys2):
    """
    This routine will take the bonding information of sys1 and copy it
    over to sys2. This routine requires that both systems have the exact
    same atoms in them. This is useful if you refragment a system and
    want to fix the bonding information.

    Args:
      sys1 (BigDFT.Systems.System): the system to copy bonding information
        from.
      sys2 (BigDFT.Systems.System): the system to copy information to.
    """
    # Check parameters
    if sys1.conmat is None:
        raise ValueError("The connectivity matrix of the first argument " +
                         "must be set")

    # Create at atom level connectivity matrix
    lookup = {}
    j = 0
    for fragid, frag in sys1.items():
        for i, at in enumerate(frag):
            lookup[(fragid, i)] = j
            j += 1

    atom_connectivity = {}
    for fid1, conn in sys1.conmat.items():
        for ati, links in enumerate(conn):
            i = lookup[(fid1, ati)]
            for (fid2, atj), v in links.items():
                j = lookup[(fid2, atj)]
                atom_connectivity[(i, j)] = v

    # Generate the matching
    atlist = []
    for fragid, frag in sys1.items():
        for at in frag:
            atlist.append(at)
    matching = sys2.compute_matching(atlist)
    rlookup = {}
    for fragid, frag in matching.items():
        for i, v in enumerate(frag):
            rlookup[matching[fragid][i]] = (fragid, i)

    # Copy over
    sys2.conmat = {}
    for fragid, frag in sys2.items():
        sys2.conmat[fragid] = []
        for i, at in enumerate(frag):
            sys2.conmat[fragid].append({})

    for (i, j), v in atom_connectivity.items():
        fragid1, at1 = rlookup[i]
        fragid2, at2 = rlookup[j]
        sys2.conmat[fragid1][at1][(fragid2, at2)] = v


def update_purity_and_bo(mapping, purity, bo, charges):
    """
    When merging fragments together, you will need to update the bond
    orders and purity values. This can be done by using
    `run_compute_purity` and `compute_bond_orders`, but this process is
    potentially slow. In this function, we use the old bond orders and
    purity values for update, which is much faster.

    Arguments:
        mapping (dict): a dictionary where the keys are the new fragments
          and the values are a list of old fragments that make up this
          new fragment.
        purity (dict): the old purity values of each fragment.
        bo (dict): the old bond orders of each fragment.
        charges (dict): the charges of each fragment (the sum of the
          number of electrons).

    Returns:
        (dict, dict): the new purity and bond order values.
    """
    # Purity loop
    new_purity = {}
    for f in mapping:
        new_purity[f] = 0
        old = mapping[f]

        # Sum the bond orders
        for o1 in old:
            for o2 in old:
                if o1 != o2:
                    new_purity[f] += bo[o1][o2]

        # Sum the un-normalized atomic purity values
        for o1 in old:
            new_purity[f] += purity[o1] * charges[o1] / 2.0

        # Normalize
        new_charge = sum([charges[x] for x in old])
        new_purity[f] *= 2.0 / new_charge

    # Bond order loop
    new_bo = {}
    for f1 in mapping:
        old1 = mapping[f1]
        new_bo[f1] = {}
        for f2 in mapping:
            old2 = mapping[f2]
            new_bo[f1][f2] = 0

            for o1 in old1:
                for o2 in old2:
                    new_bo[f1][f2] += bo[o1][o2]

    return new_purity, new_bo


def _example():
    """Example of using a system"""
    from BigDFT.IO import XYZReader
    from BigDFT.Fragments import Fragment

    safe_print("Read in some files for the fragments..")
    reader = XYZReader("SiO")
    frag1 = Fragment(xyzfile=reader)
    reader = XYZReader("Si4")
    frag2 = Fragment(xyzfile=reader)

    safe_print("Now we move on to testing the system class.")
    sys = System(frag1=frag1, frag2=frag2)
    for at in sys["frag1"]:
        safe_print(dict(at))
    for at in sys["frag2"]:
        safe_print(dict(at))
    safe_print()

    safe_print("What if we want to combine two fragments together?")
    sys["frag1"] += sys.pop("frag2")
    for at in sys["frag1"]:
        safe_print(dict(at))
    safe_print("frag2" in sys)
    safe_print()

    safe_print("What if I want to split a fragment by atom indices?")
    temp_frag = sys.pop("frag1")
    sys["frag1"], sys["frag2"] = temp_frag[0:3], temp_frag[3:]
    for at in sys["frag1"]:
        safe_print(dict(at))
    for at in sys["frag2"]:
        safe_print(dict(at))
    safe_print()


if __name__ == "__main__":
    _example()



"""A module to define typical operations that can be done on biological systems

"""
from BigDFT.Systems import System


def load(archive, serialization_version=None, options={}):
    """
    Create a Biosystem from a serialization archive.
    It extracts the archive files in a temporary directory
    and loads the related information. Also it deletes the directory
    once the load is finished.

    Args:
       archive(path): the path of the serialization archive
       serialization_version (str): version of the archive.
           Latest if not specified.

    Returns:
       BioSystem: Instance of the class, loaded from the archive
    """
    from futile.Utils import unpack_tarball
    from shutil import rmtree
    from os.path import join  # ,basename
    # import tempfile
    # import tarfile
    # import yaml
    import json
    tmpdir, files = unpack_tarball(archive, tmpdir_prefix='tmp_BioQM_archive_')
    # # creates the directory
    # tmpdir = tempfile.mkdtemp(prefix='tmp_BioQM_archive_' +
    #                           basename(archive) + '_')
    # # extract the archive
    # arch = tarfile.open(archive)
    # arch.extractall(path=tmpdir)
    # arch.close()
    # load the data
    sys = BioSystem(join(tmpdir, 'posinp.pdb'), **options)
    sys.set_qm_run(join(tmpdir, 'log.yaml'))
    for version, attrs in BioSystemSerialization.cached_attributes.items():
        if serialization_version is not None:
            desired = serialization_version
        else:
            desired = BioSystemSerialization.version
        if not _version_is_compatible(desired, version):
            continue
        for attr in attrs:
            f = open(join(tmpdir, attr+'.json'), 'r')
            value = json.load(f)
            setattr(sys, '_'+attr, value)
            f.close()
    # remove the directory
    rmtree(tmpdir)

    return sys


def _version_is_compatible(desired_version, present_version):
    desired = list(map(int, desired_version.split('.')))
    obtained = list(map(int, present_version.split('.')))
    not_ok = obtained[0] != desired[0]
    if not not_ok:
        not_ok = any([ll < m for ll, m in zip(desired[1:], obtained[1:])])
    return not not_ok


def serialize_biosystem(archive, posinp, logfile, version=None):
    """
    Serialize the biosystem associated to the provided files

    Args:
       archive (path): name of the archive to be employed for dumping
       posinp (path): pdbfile of the structure
       logfile (path): logfile of the structure
       version(str): version of the serialization
    """
    system = BioSystem(posinp)
    system.set_qm_run(logfile)
    if version is not None and _version_is_compatible('2.0', version):
        system.refragment(0.05)  # to create the atomic purities
    # system.refragment(0.025)
    serialization = BioSystemSerialization(system, version=version)
    serialization.dump(archive)


class BioSystemSerialization():
    """
    Serialization of the BioSystem class that would prevent the need to
    copy the entire matrices of the QM calculations

    Args:
        biosys (BioSystem): a instance of the Biosystem class
        version (str): the version of the desired serialization.
           If absent, the default is considered.
    """

    version = '1.1'  #: version of the serialization
    cached_attributes = {'1.0': {'purities': 'purities',
                                 'fragindices': 'frag_indices',
                                 'pairwise_BO': 'bond_orders',
                                 'cached_refragment': '_cached_refragment'},
                         '1.1': {'interactions': 'interactions'},
                         '2.0': {'atomic_purities': '_atomic_purities',
                                 'atomic_BO': 'atomic_bond_orders',
                                 'atomic_interactions': 'atomic_interactions',
                                 'cached_refragment': '_cached_refragment'}}

    def __init__(self, biosys, version=None):
        self.files = {'posinp.pdb': biosys.structure_file_path,
                      'log.yaml': biosys.logfile_path}

        self.true_version = version if version is not None else self.version

        for version, attributes in self.cached_attributes.items():
            if not _version_is_compatible(self.true_version, version):
                continue
            for attr, trueattr in attributes.items():
                setattr(self, attr, getattr(biosys, trueattr))

    def dump(self, archive):
        """
        Create an archive with the entire set of information of the
        Serialization. Such a tarfile should be such that the same
        analysis of the BioSystem is possible
        """
        from futile.Utils import create_tarball, serialize_objects
        # import json
        from BigDFT import Systems

        def if_system(sys):
            return sys.get_posinp()

        # class NumpyEncoder(json.JSONEncoder):
        #     """ Special json encoder for numpy types and System"""
        #     def default(self, obj):
        #         import numpy as np
        #         import Systems
        #         if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
        #                             np.int16, np.int32, np.int64, np.uint8,
        #                             np.uint16, np.uint32, np.uint64)):
        #             return int(obj)
        #         elif isinstance(obj, (np.float_, np.float16, np.float32,
        #                               np.float64)):
        #             return float(obj)
        #         elif isinstance(obj, (np.ndarray,)):
        #             return obj.tolist()
        #         elif isinstance(obj, (set,)):
        #             return list(obj)
        #         elif isinstance(obj, (Systems.System,)):
        #             return obj.get_posinp()
        #         return json.JSONEncoder.default(self, obj)

        dictionaries = {}
        for version, attrs in self.cached_attributes.items():
            if not _version_is_compatible(self.true_version, version):
                continue
            dictionaries.update({att+'.json': getattr(self, att)
                                 for att in attrs})

        system_encoding = {'cls': Systems.System, 'func': if_system}

        objects = serialize_objects(
                    dictionaries, extra_encoder_functions=[system_encoding])

        create_tarball(archive, self.files, objects)


class Sequence():
    """
    Initialize and define a sequence, may be amminoacidic or nitrogenous.

    Args:
       seq(Bio.Seq.Sequence) : the starting sequence
    """

    def __init__(self, seq):
        self.sequence = seq

    def __add__(self, other):
        return Sequence(self.sequence + other.sequence)

    def __radd__(self, other):
        return self if other == 0 else self.__add__(other)

    def display(self, colors=None, labels=None, axs=None, boxcount=None,
                remove_letters=[]):
        """
        Display the sequence in a easy-to-read form

        Args:
            boxcount(int): the number of boxes per row
            remove_letters (list): list of letter positions that will be
                removed
        """
        def formatfun(x):
            return ("{: ^"+str(llen)+"}").format(x)
        # Handle optional axis
        if axs is None:
            import matplotlib.pyplot as plt
            fig, axs = plt.subplots()
        axs.axis('off')
        # bbox = axs.get_window_extent()

        # Handle the sizes based on the presence of labels
        if labels:
            yoffset = 24
            boxcount = 30 if boxcount is None else boxcount
            llen = max([len(x) for x in labels if x is not None])
            xoffset = 3*llen + 5
        else:
            xoffset = 7
            yoffset = 14
            boxcount = 40 if boxcount is None else boxcount

        # Handle the matplotlib options
        matargs = {"size": 18, "va": "center", "ha": "center",
                   "multialignment": "left"}
        matargs["fontdict"] = {"family": "monospace", "weight": "normal"}

        xcoord = 0
        ycoord = 100
        removed = False
        for i, letter in enumerate(str(self.sequence)):
            if i in remove_letters:
                if not removed:
                    matargs['bbox'] = {"linewidth": 0, 'fill': False}
                    axs.text(0.01 * xcoord, 0.01 * ycoord, "-",
                             **matargs)
                    # Update coordinates
                    xcoord += xoffset
                    removed = True
                continue
            removed = False
            label = None if labels is None or labels[i] is None else labels[i]
            fc = "none" if colors is None or colors[i] is None else colors[i]

            if isinstance(fc, tuple):
                lc = fc[1]
                fc = fc[0]
            else:
                lc = fc

            if label:
                matargs["bbox"] = {"fc": lc, "linewidth": 0}
                axs.text(0.01 * xcoord, 0.01 * ycoord + 0.05, formatfun(""),
                         **matargs)
                matargs["bbox"] = {"fc": fc, "linewidth": 0}
                axs.text(0.01 * xcoord, 0.01 * ycoord - 0.05, formatfun(""),
                         **matargs)

                matargs["bbox"] = {"fill": False, "linewidth": 1,
                                   "edgecolor": "k"}
                axs.text(0.01 * xcoord, 0.01 * ycoord, formatfun(label) +
                         "\n" + formatfun(letter), **matargs)
            else:
                matargs["bbox"] = {"fc": fc, "linewidth": 1, "edgecolor": "k"}
                axs.text(0.01 * xcoord, 0.01 * ycoord, letter, **matargs)

            # Update coordinates
            xcoord += xoffset
            if xcoord > boxcount * xoffset:
                xcoord = 0
                ycoord -= yoffset
        return axs


def order_chains(fragments):
    from Bio.PDB.Polypeptide import three_to_one as letter
    chains = {}
    for f in fragments:
        ch, res, num = construct_frag_tuple(f)
        resnum = ':'.join([res, str(num)])
        try:
            lett = letter(res)
        except KeyError:
            lett = 'X'
        chains.setdefault(ch, {})[resnum] = lett
    return chains


def residue_to_chains(chdict, residues):
    chlett = {}
    for ires, res in enumerate(residues):
        ch, resname, num = construct_res_tuple(res)
        resnum = ':'.join([resname, str(num)])
        lett = chdict[ch].pop(resnum)
        if lett != 'X':
            chlett.setdefault(ch, {})[num] = (lett, ires)
    assert all(len(ch) == 0 for ch in chdict.values())
    return chlett


def chains_to_sequences(chlett):
    from Bio.Seq import Seq
    from Bio.Alphabet import ProteinAlphabet
    seqs = []
    chains_to_residues = []
    for ch in sorted(chlett.keys()):
        chain = chlett[ch]
        strseq = ''.join([chain[i+1][0] for i in range(len(chain))])
        seqs.append(Sequence(Seq(strseq, alphabet=ProteinAlphabet)))
        chains_to_residues.append([chain[i+1][1] for i in range(len(chain))])
    return seqs, chains_to_residues


def pdb_sequences(filename):
    """
    Sequences of the chains of the proteins, in the order of the residue_list

    Args:
       filename (str): protein file, in pdb format

    Returns:
       list: strings of the protein sequence in the FASTA sequence order
    """
    import Bio.SeqIO as S
    return [Sequence(rec.seq) for rec in S.parse(filename, format='pdb-atom')]


def read_structure(filename, name='protein'):
    """
    Extract a biological structure from a PDB file. Use the package Bio.PDB

    Args:
       filename (str): path to the file of the structure
       name (str): the name of the structure

    Returns:
       (Bio.PDB.Structure) A structure instance
    """
    import Bio.PDB as pdb
    parser = pdb.PDBParser()
    return parser.get_structure(name, filename)


def residue_list(structure, system, strict=True):
    """
    Identify the residues of the sequence in the fragment of the system

    Args:
       structure (structure): A structure read from the PDB parser of Bio.PDB
       system (BigDFT.Systems.System): a system coming from BigDFT package
       strict (bool): define if the identification has to be performend via
          centroids control or just by the residue name

    Returns:
       tuple: tuple of list the fragments of the `system` in the order of the
           residues of the structure. Some of the fragments may be missing as
           they may have been merged of not associated to elements of the
           sequence. The second list is the list of the ordered residues.
    """
    from BigDFT.Atoms import AU_to_A
    import numpy as np
    reslist = []
    centroids_dict = {frag: system[frag].centroid * AU_to_A for frag in system}
    original_residues = []
    for residue in structure.get_residues():
        restuple = construct_res_tuple(residue)
        centroid = sum([a.coord for a in residue.get_atoms()]) / len(residue)
        found = False
        for frag in centroids_dict:
            fragtuple = construct_frag_tuple(frag)
            if not strict:
                found = restuple == fragtuple
            else:
                found = np.linalg.norm(centroids_dict[frag] - centroid) < 1.e-4
            if found:
                reslist.append(frag)
                centroids_dict.pop(frag)
                break
        if not found:
            reslist.append(None)
        original_residues.append(residue)
    return reslist, original_residues


def sequences_to_residues(reslist, chains):
    """
    Identify for each element of the chains which are the residues that may be
    associated to them.

    Args:
       reslist(list): the list of all the residues of the system
       chains (list): a list containing the sequences of the system

    Returns:
       list: a list of length of the chains indicating the lookup array for the
           sequence
    """
    lookup = [[None for k in c.sequence] for c in chains]
    chain_letters = []
    for ires, res in enumerate(reslist):
        chain_id = res.full_id[2]
        if chain_id not in chain_letters:
            chain_letters.append(chain_id)
        ch = chain_letters.index(chain_id)
        pos = res.full_id[3][1] - 1
        if ch < len(lookup) and ch >= 0 and pos < len(lookup[ch]) and pos >= 0:
            lookup[ch][pos] = ires
    return lookup


class Graph():
    """
    A Graph representation of a System or a subportion of it.

    Identify the connectivity of the system given a threshold, possibly
    restricted to a subset of fragments
    Args:
       cutoff (float) : the threshold of the cumulative fragment bond order
        that is applied to define the connectivity
       restrict_to (list): list of the fragment id to which calculate
        the connectivity
    """

    def __init__(self, fragkeys, bond_orders, frag_labels, cutoff=0.01,
                 restrict_to=None):
        import numpy as np
        self.connectivity_matrix = graph_bond(
            fragkeys, cutoff, bond_orders, restrict_to=restrict_to)

        self.nw, self.target_nodes, self.target_edges = get_BO_network(
                                                    self.connectivity_matrix.T)

        self.labels = {i: l for i, l in enumerate(frag_labels)
                       if i in self.target_nodes}

        if restrict_to is not None:
            self.restrict_nodes = [fragkeys.index(f) for f in restrict_to]
        else:
            self.restrict_nodes = None

        frags = fragkeys

        self.edge_weights = []
        for i, j in self.nw.edges:  # self.target_edges:
            wgt = bond_orders[frags[i]][frags[j]] if i != j else -1
            self.edge_weights.append(wgt)

        self.edge_weights = np.array(self.edge_weights)
        wh = np.where(self.edge_weights == -1)
        mx = np.max(self.edge_weights)
        self.edge_weights[wh] = mx

    def display(self, colors=None, edge_colorcode='Reds', axs=None,
                edge_labels=None, **kwargs):
        """
        Draw the graph in the Kamada Kawai representation.
        Employ colors if present.

        Args:
           colors (list): the list of the colors that have to be employed for
              all the graph nodes. If the target nodes are present only the
              selected nodes are employed
           edge_colors (list): the list of the colors that have to be employed
              for each of the edges of the node
           axs (matplotlib.axis): the axis to draw on (optional)
           edge_labels (dict): a dictionary from tuples of node labels to
             a given label.
           kwargs (dict): any other argument you wish to pass to networkx.

        """
        from networkx import kamada_kawai_layout, draw_networkx
        from networkx import draw_networkx_edge_labels, draw_networkx_nodes

        # Handle optional axis
        if axs is None:
            import matplotlib.pyplot as plt
            fig, axs = plt.subplots(1, 1, figsize=(18, 12))

        # Setup colors
        node_color = 'r' if colors is None else [
            colors[i] for i in self.target_nodes]

        # Handle kwargs
        args = {"labels": self.labels, "node_size": 800,
                "font_size": 10, "font_family": 'serif', "linewidths": 4,
                "font_weight": 'bold', "alpha": 0.95,
                "ax": axs, "edge_color": 'k'}
        args.update(kwargs)

        layout = kamada_kawai_layout(self.nw)

        if self.restrict_nodes is not None:
            node_color = 'r' if colors is None else [
                colors[i] for i in self.restrict_nodes]
            draw_networkx_nodes(self.nw, layout, node_shape='s',
                                node_color=node_color,
                                nodelist=self.restrict_nodes, **args)

        node_color = 'r' if colors is None else [
            colors[i] for i in self.target_nodes]
        draw_networkx(self.nw, layout, node_color=node_color,
                      nodelist=self.target_nodes, **args)

        if edge_labels is not None:
            draw_networkx_edge_labels(self.nw, layout,
                                      nodelist=self.target_nodes,
                                      node_color=node_color,
                                      edge_labels=edge_labels, **args)


def construct_frag_tuple(frag):
    """
    Build the fragment tuple given the fragment name.

    Args:
        frag (str): the fragment name

    Returns:
        tuple: the (chain,fragname,fragno) tuple name
    """
    ch, resnum = frag.split('-')
    res, num = resnum.split(':')

    return ch, res, int(num)


def construct_res_tuple(res):
    """
    Build the BigDFT fragment tuple given the residue of the structure

    Args:
       res(Residue): A residue Class on the Biopython package

    Returns:
       tuple: the (chain,fragname,fragno) tuple
    """
    chain = res.full_id[2]
    if len(chain.lstrip(' ')) == 0:
        chain = 'A'
    resname = res.resname
    position = res.full_id[3][1]
    return chain, resname, position


def residue_name(res):
    """
    Name of the residue in the fragment specification of PyBigDFT

    Args:
       res(Residue): A residue Class on the Biopython package
    Returns:
       str: name of the fragment
    """
    chain, resname, position = construct_res_tuple(res)
    return ':'.join(('-'.join((chain, resname)), str(position)))


def biosystem_fragmentation(sys, structure):
    """
    Refragment the system names according to the residues name of the
    structure from the biopython module.

    Args:
       sys (System): and instance of the BioSystem class, fragmented in a
       traditional way
       structure (Bio.Structure): a structure coming from the BioPython module

    Returns:
       System: a system with a fragmentation that is dependent of the
          chain of the molecules
    """
    from BigDFT.Fragments import Fragment
    from BigDFT.Systems import System
    from warnings import warn
    # from Atom import Atom
    # newsys = [at for at in sys.get_posinp()['positions']]
    # units = sys.get_posinp()['units']
    from BigDFT.Atoms import AU_to_A
    # from Fragments import system_from_dict_positions
    mproat = [{at.element: list(at.coord/AU_to_A), 'atom': at,
               'frag': residue_name(at.get_parent()),
               'chain':
               at.full_id[2] if len(at.full_id[2].lstrip(' ')) > 0 else 'A'}
              for at in structure.get_atoms()]
    try:
        lookup = sys.compute_matching(atlist=mproat)
    except ValueError as e:
        lookup = sys.compute_matching(atlist=mproat, check_matching=False)
        warn('Some atoms were not matched' + str(e))
    newsys = System()
    for frag in sys:
        for at, jat in zip(sys[frag], lookup[frag]):
            # reat = mproat[jat]['atom']
            # newname = residue_name(reat.get_parent())
            newname = mproat[jat]['frag']
            if frag not in newname:
                newname = '-'.join([mproat[jat]['chain'], frag])
            # if newname not in newsys:
            #     newsys[newname] = Fragment()
            newsys.setdefault(newname, Fragment()).append(at)
    return newsys


def sequences_to_fragments(chains_to_residues, frag_to_residues):
    """
    Defines the lookup array of the sequences to the fragments

    Args:
       chains_to_residues(list): lookup of the chains to the list of recogized
             residues
       frag_to_residues (list): name of the associated fragments in the order
             of residues list

    Returns:
       list: array of the name of the fragments in the same representation of
           chain_to_residues
    """
    from copy import deepcopy
    lookup = deepcopy(chains_to_residues)
    for c in lookup:
        for i, ires in enumerate(c):
            if ires is None:
                continue
            frag = frag_to_residues[ires]
            c[i] = frag
    return lookup


def _aggregate_values(fragments, data, view, op=sum):
    newdata = [d for d in data]
    for multiple_frag in view.values():
        datasum = op(data[fragments.index(f)] for f in multiple_frag)
        for f in multiple_frag:
            newdata[fragments.index(f)] = datasum
    return newdata


class BioSystem(System):
    """
    Initialize a System for a Biological analysis.
    Retrieve, from the initial file, the basic fragments, the system sequence,
    their mutual interactions and the informations about the purity and the
    observables

    Args:
       filename (str): a pdb file with fragment information inside
       sys (System): an initial system, useful for a custom refragmentation.
            If present, overrides the association written in the pdbfile.
            It is user's responibility to ensure coherency of the total
            number of atoms.
        sequence_from_fragments (bool): method to determine the sequences.
            If True, the sequence are determined by the fragment names.
            It is user's responsibility to provide a inputfile that has a
            set of residues compatible with the fragmentation.
        disable_warnings (bool): disable the BioPython warnings if true
        structure (Bio.Structure): a system structure that is meant to replace
            the one genrated by the file parser. Useful for corrupted pdb
            files. To be used with `sequence from_fragments`
    """
    def __init__(self, filename, sys=None, structure=None,
                 sequence_from_fragments=False, disable_warnings=False):
        from BigDFT.IO import read_pdb
        from os import path
        import warnings
        # disable warnings if asked for
        if disable_warnings:
            from Bio import BiopythonWarning
            warnings.simplefilter('ignore', BiopythonWarning)
        # path of the file
        self.structure_file_path = path.abspath(filename)
        if structure is None:
            # read structure from the Bio.PDB format
            # (can replace custom read_pdb)
            self.structure = read_structure(filename)
        else:
            self.structure = structure

        if sys is None:
            # read pdb from the custom routine
            with open(filename) as ifile:
                sys = read_pdb(ifile)
            # reorganize the fragmentation
            newsys = biosystem_fragmentation(sys, self.structure)
        else:
            newsys = sys

        System.__init__(self, **newsys.dict())

        self._residue_identification(sequence_from_fragments)

        self._sequence_identification(sequence_from_fragments)

        # once the system is identified proceed with the association
        self._fragment_identification()

    def _residue_identification(self, sequence_from_fragments):
        # match the system's fragment into the amminoacids of he structure
        self.frag_to_residues, self.residues = residue_list(
                                        self.structure, self,
                                        strict=not sequence_from_fragments)

    def _sequence_identification(self, sequence_from_fragments):
        # identify sequences
        if sequence_from_fragments:
            chdict = order_chains(self.keys())
            chlett = residue_to_chains(chdict, self.residues)
            self.chains, self.chains_to_residues = chains_to_sequences(chlett)
        else:
            self.chains = pdb_sequences(self.structure_file_path)  # filename)
            # convert the element of the sequences to residues (if possible)
            self.chains_to_residues = sequences_to_residues(
                self.residues, self.chains)
        # associate to each of the chain elements the corresponding fragment
        # names
        self.sequences_to_fragments = sequences_to_fragments(
                              self.chains_to_residues,
                              self.frag_to_residues)

    def _fragment_identification(self):
        # store the fragment keys in a given order
        self.fragment_names = list(self.keys())
        # obtain the list of the letters of each fragment (if possible)
        self.fragment_letters = fragment_letters(
                        self.fragment_names,
                        [seq.sequence for seq in self.chains],
                        self.frag_to_residues,  self.chains_to_residues)
        # identify the fragments that are unmatched either in the residues or
        # in the sequence
        self.unmatched_fragments = [frag for frag, lett in zip(
            self.fragment_names, self.fragment_letters) if lett == 'X' or
            lett == 'None' or lett is None]

    def set_qm_run(self, log):
        """
        Associate the system to a QM run that has been performed by BigDFT.
        Such run assumes that the positions of the system atoms are consistent
        with the PDB

        Args:
           log (Logfiles.Logfile): A Logfile of the BigDFT run
        """
        import numpy as np
        from os import path
        from BigDFT.Logfiles import Logfile
        from BigDFT.Systems import system_from_log
        # path of the logfile
        self.logfile_path = path.abspath(log)
        self.logfile = Logfile(log)
        # TODO: We should insert the compute_matching procedure which checks if
        # the systems are coherent (all the atoms are OK)
        self.set_logfile_info(self.logfile)
        self.fragment_charges = np.array(
            [self[frag].q0[0] for frag in self.fragment_names])
        self.fragment_dipoles = np.array([self[frag].d1()
                                          for frag in self.fragment_names])
        self.fragment_dipole_norms = np.array(
            [np.linalg.norm(d1) for d1 in self.fragment_dipoles])
        atomic_system = system_from_log(self.logfile,
                                        fragmentation='atomic')
        rename_mapping = {'ATOM:'+str(i): i for i in range(len(atomic_system))}
        self.atomic_system = atomic_system.rename_fragments(rename_mapping)
        lookup = self.compute_matching(
            [x[0] for x in self.atomic_system.values()])
        self.atomic_lookup = {x: [z for z in y] for x, y in
                              lookup.items()}
        self._cached_refragment = {}

    def refragment(self, cutoff=0.05, view=None):
        """
        Identify the system's fragmentation and represent it in terms of the
        systems' residues.
        Employs the user defined cutoff to find the correct fragmentation

        Args:
            cutoff (float): cutoff employed for the refragmentation
            view (dict): the view to be imposed as a starting
                fragmentation in the system
        """
        from BigDFT import Systems

        data = self._cached_refragment.get(str(cutoff))
        if data is not None:
            if isinstance(data, (list, tuple)):
                if not isinstance(data[0], Systems.System):
                    pos = data[0]['positions']
                    unt = data[0]['units']
                    data[0] = Systems.system_from_dict_positions(pos,
                                                                 units=unt)
                remapping = {frag: frag.split('+') for frag in data[0]}
            else:
                remapping = data
            return remapping
        fw = Systems.FragmentView(self.purities, self.bond_orders,
                                  charges={k: f.nel for k, f in self.items()})
        if view is None:
            new_view = fw
            sys = Systems.System(self)
        else:
            new_view = fw.refragment(view)
            sys = self.reform_superunits(view)
        remapping = self.bigdft_tool.auto_fragment(sys, new_view, cutoff,
                                                   verbose=True,
                                                   criteria="bondorder")
        # the mapping has to be reworked in terms of a view
        # to prevent renaming of the original fragments
        if view is not None:
            remap = {}
            for frag, lst in remapping.items():
                if frag in view and lst == view[frag]:
                    remap[frag] = lst
                    continue
                remap[frag] = []
                for orig_frag in view:
                    if all(f in frag.split('+') for f in orig_frag.split('+')):
                        remap[frag] += view[orig_frag]
            remapping = remap
        self._cached_refragment[str(cutoff)] = remapping
        return remapping

    def chessboards(self, cutoffs, initial_view=None):
        """
        Define a coherent group of fragmentations that are provided
        from the loosest to tightest cutoffs.
        The views of the fragmentations are defined such as the tighest
        cutoffs always contain regrouping of fragments from the loosest

        Args:
            cutoffs (list): list of floating point numbers of the desired
                cutoffs values
            initial_view (dict): mapping of the initial fragmentation that will
                be imposed strating from the loosest cutoff
        Returns:
            dict: dictionary of the view. the values of the dictionary are
                length-1 lists such that subsequent views (for instance found
                when instantiating a population) can be appended
        """
        coff = sorted(cutoffs, reverse=True)
        cb = {}
        current_view = None if initial_view is None else initial_view.copy()
        self._cached_refragment = {}  # remove the caching for refragmenting
        for c in coff:
            current_view = self.refragment(c, view=current_view)
            cb[c] = [current_view]
        return cb

    @property
    def fragment_ids(self):
        if not hasattr(self, '_fragment_ids'):
            self._fragment_ids = {}
            for frag in self.fragment_names:
                ifrag = 0
                found = False
                for ch in self.sequences_to_fragments:
                    if frag in ch:
                        jfrag = ch.index(frag)
                        found = True
                        break
                    ifrag += len(ch)
                if not found:
                    jfrag = self.unmatched_fragments.index(frag)
                self._fragment_ids[frag] = jfrag + ifrag
        return self._fragment_ids

    def _fragment_labels(self, view=None):
        label = {}
        if view is None:
            view = {f: [f] for f in self.fragment_names}
        for lf in view.values():
            lb = min([self.fragment_ids[f] for f in lf])
            for f in lf:
                if self.fragment_ids[f] == lb:
                    relb = lb
                    for ch in self.chains_to_residues:
                        if relb >= len(ch):
                            relb -= len(ch)
                    label[f] = str(relb+1)
                else:
                    label[f] = '   '
        return label

    @property
    def purities(self):
        from BigDFT.PostProcessing import superunits_purities
        if not hasattr(self, '_atomic_purities') and \
           not hasattr(self, '_purities'):
            self._atomic_purities = self.bigdft_tool.run_compute_purity(
                             self.atomic_system, self.logfile,
                             kxs=self.ks_matrix,
                             frag_indices=self.atomic_frag_indices)
        if not hasattr(self, '_purities'):
            atomic_charges = {iat: at.nel for iat, at in
                              self.atomic_system.items()}
            fragment_charges = {frag: f.nel for frag, f in
                                self.items()}
            self._purities = superunits_purities(self.atomic_bond_orders,
                                                 self._atomic_purities,
                                                 atomic_charges,
                                                 self.atomic_lookup,
                                                 fragment_charges)
        return self._purities

    @property
    def fragment_purities(self):
        """
        Define the quality of the fragmentation that is provided in the system
        """
        return [abs(self.purities[frag]) for frag in self.fragment_names]

    @property
    def ks_matrix(self):
        if not hasattr(self, '_kxs'):
            self._kxs = self.bigdft_tool.get_matrix_kxs(self.logfile)
        return self._kxs

    @property
    def sinvh_matrix(self):
        if not hasattr(self, '_sinvh'):
            self._sinvh = self.bigdft_tool.get_matrix_sinvh(self.logfile)
        return self._sinvh

    @property
    def atomic_frag_indices(self):
        if not hasattr(self, '_atomic_fragindices'):
            self._atomic_fragindices = self.bigdft_tool.get_frag_indices(
                self.atomic_system, self.logfile)
        return self._atomic_fragindices

    @property
    def frag_indices(self):
        if not hasattr(self, '_fragindices'):
            self._fragindices = self.bigdft_tool.get_frag_indices(
                self, self.logfile)
        return self._fragindices

    @property
    def bigdft_tool(self):
        from BigDFT.PostProcessing import BigDFTool
        if not hasattr(self, '_btool'):
            from os import environ as env
            if 'BIGDFT_MPIRUN' not in env:
                # to be customized
                env['BIGDFT_MPIRUN'] = 'OMP_NUM_THREADS=2 mpirun -np 2'
            self._btool = BigDFTool()
        return self._btool

    def d3(self):
        from ase.calculators.dftd3 import DFTD3
        from os import environ
        from os.path import join
        from futile.Utils import unique_filename
        if 'ASE_DFTD3_COMMAND' not in environ:
            environ["ASE_DFTD3_COMMAND"] = join(environ["BIGDFT_ROOT"],
                                                "dftd3")
        return DFTD3(xc='pbe', grad=False, damping="bj",
                     label=unique_filename(prefix='ased3_'))

    @property
    def atomic_bond_orders(self):
        if not hasattr(self, '_atomic_BO'):
            atoms = [i for i in range(len(self.atomic_system))]
            self._atomic_BO = self.bigdft_tool.fragment_bond_order(
                       self.atomic_system, atoms, atoms,
                       self.logfile, kxs=self.ks_matrix,
                       frag_indices=self.atomic_frag_indices)
        return self._atomic_BO

    @property
    def bond_orders(self):
        from BigDFT.PostProcessing import superunits_quadratic_quantities
        if not hasattr(self, '_pairwise_BO'):
            self._pairwise_BO = superunits_quadratic_quantities(
                                self.atomic_bond_orders, self.atomic_lookup)
            # self._pairwise_BO = self.bigdft_tool.fragment_bond_order(
            #            self, self.fragment_names, self.fragment_names,
            #            self.logfile, kxs=self.ks_matrix,
            #            frag_indices=self.frag_indices)
        return self._pairwise_BO

    @property
    def atomic_interactions(self):
        if not hasattr(self, '_atomic_interactions'):
            atoms = [i for i in range(len(self.atomic_system))]
            self._atomic_interactions = \
                self.bigdft_tool.fragment_interaction_energy(
                         self.atomic_system, atoms, atoms,
                         self.logfile, kxs=self.ks_matrix,
                         frag_indices=self.atomic_frag_indices,
                         sinvh=self.sinvh_matrix)
        return self._atomic_interactions

    @property
    def interactions(self):
        from BigDFT.PostProcessing import superunits_quadratic_quantities
        if not hasattr(self, '_atomic_interactions') and \
           not hasattr(self, '_interactions'):
            self._interactions = self.bigdft_tool.fragment_interaction_energy(
                     self, self.fragment_names, self.fragment_names,
                     self.logfile, kxs=self.ks_matrix,
                     frag_indices=self.frag_indices,
                     sinvh=self.sinvh_matrix)
        elif not hasattr(self, '_interactions'):
            self._interactions = superunits_quadratic_quantities(
                                 self.atomic_interactions,
                                 self.atomic_lookup)
        return self._interactions

    def fragment_distances(self, target_fragments):
        """
        Calculate the distance between each of the fragment and
        a list of target fragments.

        Args:
            target_fragments (list): Id of the fragments to focus on

        Returns:
            array-like: list of the distances in the order of fragment_names
        """
        from BigDFT.Fragments import pairwise_distance
        from numpy import array

        distances = []
        for f in self.fragment_names:
            distances.append(
                min([pairwise_distance(self[f], self[g])
                     for g in target_fragments]))
        return array(distances)

    def fragment_interaction_strengths(self, target_fragments,
                                       criteria='hamiltonian'):
        """
        Fragment Bond Order between the different fragments of the system,
        possibly specified on target regions.

        Args:
           target_fragments (list): Id of the fragments to focus on
           criteria (str): may be 'hamiltonian', 'electrostatic',
               or 'bond_order'

        Returns:
           array-like: list of the strengths of the interactions of the
                fragments wrt to the target fragments.
                The interaction of the target fragments with themselves is put
                to the maximum value of the other interactions in order
                to avoid scale misbehaviour
        """
        ints = {'bond_order': 'bond_orders',
                'hamiltonian': 'interactions',
                'electrostatic': 'electrostatic_interactions'}
        return interaction_strengths(self.fragment_names, target_fragments,
                                     getattr(self, ints[criteria]))

    def d3PBE_energy(self):
        """
        Full dispersion energy of the system

        Returns:
            float: the d3 PBE energy for the PBE functional, in Hartree
        """
        energy = self.ase_potential_energy(self.d3())
        return energy

    def display_sequences(self, fig=None, labeldict=None, boxcount=None,
                          with_fragment_labels=False, remove_letters=None,
                          **kwargs):
        """
        Display the full sequencies of the system.
        Employ the given field to define the color dict if available

        Args:
           field_vals(list) : values of the field to decide
               the colors of the keys
           labeldict (dict): an optional dictionary mapping fragment ids
               to strings that will be printed on each sequence square.
           colordict (dict) :  the dictionary of the keys,
               and the corresponding colors. Overrides field_vals if present
           fig (matplotlib.pyplot.Figure): figure on which to put the sequence
               into
           with_fragment_labels (bool): a boolean useful to represent the
               fragment numbers. Can be combined with a view to visualize the
               QM fragments. Provides a default labeldict.
           remove_letters (list): list of the letters to be removed from
               the display. In elements of the chains.
           **kwargs: arguments to be passed to display of the colordict
        Returns:
           matplotlib.pyplot.Figure: the figure in which the sequences were
               displayed
        """
        # import matplotlib.pyplot as plt
        colord = self.colordict(**kwargs)
        if with_fragment_labels and labeldict is None:
            labeldict = self._fragment_labels(kwargs.get('view'))

        if labeldict is not None and 'colordict' not in kwargs:
            colord = self._colordict_for_labelling(colord, **kwargs)

        # if fig is None:
        #     fig = plt.figure()
        ifig = 1
        # nrows = len(self.chains)
        for i, (seq, lookup) in enumerate(zip(self.chains,
                                              self.sequences_to_fragments)):
            colorletters = [colord.get(frag) for frag in lookup]
            if remove_letters is None:
                remove_chain_letters = []
            else:
                remove_chain_letters = remove_letters[i]
            if labeldict is not None:
                labelletters = [labeldict.get(frag) for frag in lookup]
            else:
                labelletters = None
            # ax = fig.add_subplot(nrows, 1, ifig)
            # ax =
            seq.display(colors=colorletters, labels=labelletters,
                        boxcount=boxcount, remove_letters=remove_chain_letters)
            ifig += 1
        return fig

    def display_graph(self, restrict_to=None, **kwargs):
        """
        Display a graph view of the system

        Args:
           restrict_to (list): list of fragmens to which the graph
                 has to be restricted
           **kwargs: colordict args
        """
        G = Graph(self.fragment_names, self.bond_orders, self.fragment_letters,
                  restrict_to=restrict_to)

        colord = self.colordict(**kwargs)

        G.display(colors=[colord[i] for i in self.fragment_names])

    def display(self, cartoon=False, **kwargs):

        System.display(self, colordict=self.colordict(**kwargs),
                       cartoon=cartoon)

    def _colordict_for_labelling(self, colord, **kwargs):
        from futile.Utils import kw_pop
        if 'color_by' not in kwargs and 'field_vals' not in kwargs:
            fragment_colors = colord
            cd = {name: ('white', fragment_colors.get(name, 'white'))
                  for name in colord}
        else:
            new_kw, tt = kw_pop('color_by', None, **kwargs)
            new_kw, tt = kw_pop('field_vals', None, **new_kw)
            fragment_colors = self.colordict(**new_kw)
            cd = {name: (fc, fragment_colors.get(name, 'white'))
                  for name, fc in colord.items()}
        return cd

    def refragmentation_colordict(self, refrag_keys):
        """
        Define a colordict that associates the fragment which go together
        with the same color and put to white all the previously pure fragments.

        Args:
           refrag_keys (list): Keys of the refragmented system

        Returns:
           dict: the dictionary of the colors (equal color means same fragment)
        """
        colordict = self.colordict()
        rekeys = {}
        for frag in refrag_keys:
            # if isinstance(frag, tuple):
            if '+' in frag:
                allfrags = frag.split('+')
                recolor = colordict[allfrags[0]]
                rekeys[allfrags[0]] = []
                for refrag in allfrags[1:]:
                    colordict[refrag] = recolor
                    rekeys[allfrags[0]].append(refrag)
            else:
                colordict[frag] = 'white'
        cc = self.colordict(keys=rekeys.keys())
        for leadfrag in rekeys:
            for refrag in rekeys[leadfrag]:
                cc[refrag] = cc[leadfrag]
        return cc

    def colordict(self, color_by=None, keys=None, field_vals=None,
                  colordict=None, colorcode='html', highlight=None,
                  view=None):
        """
        Define the colordict to quantify the fragments
        according to the method

        Args:
           keys (list): keys of the fragment names
           color_by (str): can be 'charge', 'dipole', 'purity'
           field_vals(list) : values of the field to decide
                 the colors of the keys
           colordict (dict) :  the dictionary of the keys,
                 and the corresponding colors. Overrides field_vals if present
           colorcode (str): the string of the colorcode. can be 'html' or 'rgb'
                if field_dict is absent, otherwise it represents
                the colormap of matplotlib.
                Default is 'seismic' for diverging values
                (i.e. field_dict has negative data), otherwise 'Reds'
            highlight (list) : color in red only the fragments which
                are indicated in the list
            view (dict): the fragmentation of the system to aggregate
                the field_vals to
        """
        from BigDFT import Visualization as V
        if keys is None:
            keys = self.fragment_names
        if colorcode == 'html':
            colorcode = self._colorcode(criteria=color_by)
        if field_vals is None:
            if color_by is not None:
                field_vals = self.fragment_values(criteria=color_by, view=view)
            elif view is not None and colordict is None:
                return self.refragmentation_colordict(view)
        elif view is not None:
            field_vals = _aggregate_values(keys, field_vals, view)
        if highlight is not None:
            colord = {
                frag: 'red' if frag in highlight else 'white' for frag in self}
        elif colordict is None:
            colord = V.get_colordict(keys=keys, field_vals=field_vals,
                                     colorcode=colorcode)
        else:
            colord = colordict
        return colord

    def _colorcode(self, criteria=None):
        """
        Retrieve the colorcode to be used for the numpy cmap
        This is a commodity function made to make more intentional
        some treatment.

        Args:
           criteria (str): can can be 'charge', 'dipole', 'purity'

        Returns:
           str: matplotlib cmap to be used
        """
        if criteria is None:
            return 'html'
        if criteria == 'charge':
            colorcode = 'seismic'
        elif criteria == 'dipole':
            colorcode = 'Greens'
        elif criteria == 'purity':
            colorcode = 'Greys'
        else:
            colorcode = 'html'
        return colorcode

    def fragment_values(self, criteria=None, view=None):
        """
        Retrieve the fragment values that are associated to a give criteria.
        This is a commodity function made to make more intentional some data.

        Args:
           criteria (str): can can be 'charge', 'dipole', 'purity'
           view (dict): the dictionary of the fragment view. If present,
               aggregate the values accordingly.

        Returns:
           list: in order of the fragment names.
        """
        from BigDFT.Systems import FragmentView
        from numpy.linalg import norm
        if criteria is None:
            return None
        field_vals = None
        if criteria == 'charge':
            if view is None:
                field_vals = self.fragment_charges
            else:
                field_vals = _aggregate_values(self.fragment_names,
                                               self.fragment_charges, view)
        elif criteria == 'dipole':
            if view is None:
                field_vals = self.fragment_dipole_norms
            else:
                aggregated_dipoles = _aggregate_values(self.fragment_names,
                                                       self.fragment_dipoles,
                                                       view)
                field_vals = map(norm, aggregated_dipoles)
        elif criteria == 'purity':
            field_vals = self.fragment_purities
            if view is not None:
                fw = FragmentView(self.purities, self.bond_orders,
                                  {f: ff.nel for f, ff in self.items()})
                new_fw = fw.refragment(view)
                for frag, pv in new_fw.purities.items():
                    for f in view[frag]:
                        field_vals[self.fragment_names.index(f)] = abs(pv)
        return field_vals

    def fragment_scatterplot(self, errors=None, **kwargs):
        """
        Defines the scatterplot to have a look to in order to interpret
        sequence data

        Args:
           errors (list): the errors of the field_vals in the order
               of fragment_names
           **kwargs: the same arguments of the `py:meth:colordict` method

        Returns:
           module: reference to the `matplotlib.pyplot` module
        """
        import matplotlib.pyplot as plt
        colors = self.colordict(**kwargs)
        field_vals = kwargs.get('field_vals')
        view = kwargs.get('view')
        if field_vals is None:
            field_vals = self.fragment_values(criteria=kwargs.get('color_by'),
                                              view=view)
        elif view is not None:
            field_vals = _aggregate_values(self.fragment_names, field_vals,
                                           view)
        for i, (f, p) in enumerate(zip(self.fragment_names, field_vals)):
            j = self.fragment_ids[f]
            plt.scatter(j, p, color=colors[f])
        if errors is not None:
            for iname, (y, e) in enumerate(zip(field_vals, errors)):
                x = self.fragment_ids[self.fragment_names[iname]]
                plt.errorbar(x, y, e, fmt='none', elinewidth=0.5, ecolor='k')
        plt.xlabel('Fragment ID')
        return plt

    def represent(self, errors=None, with3d=False, with_fragment_labels=False,
                  boxcount=None, remove_letters=None, **kwargs):
        """
        Represent all the data of the systems, both in with the sequence
        and with the scatterplot. Useful only when there is data to be
        represented.

        Args:
           with3d (bool): if True represent the 3d vision of the system
           errors (list): the errors of the field_vals in the order
               of fragment_names
           **kwargs: the same arguments of the `py:meth:colordict` method

        Returns:
           module: reference to the `matplotlib.pyplot` module
        """
        plt = self.fragment_scatterplot(errors=errors, **kwargs)
        self.display_sequences(with_fragment_labels=with_fragment_labels,
                               boxcount=boxcount,
                               remove_letters=remove_letters,
                               **kwargs)
        if with3d:
            self.display(**kwargs)
        return plt

    def unmatched_frag_info(self):
        """
        Retrieve the information about the fragments which have not been
        matched inside the sequence

        Returns:
           dict: a Dictionary of the fragment that have not been recognized
        """
        from BigDFT.Atoms import AU_to_A
        info = {'expected': {}, 'found': {}}
        expected = []
        for inum, (frag, residue) in enumerate(zip(self.frag_to_residues,
                                                   self.residues)):
            if frag is not None:
                letter = self.fragment_letters[self.fragment_names.index(frag)]
            if frag is not None and letter is not None:
                continue  # fragment has been matched
            centroid = sum([a.coord for a in residue.get_atoms()])/len(residue)
            info['expected'][residue_name(residue)] = {
                 'centroid': list(centroid), 'nat': len(residue),
                 'residue': residue}
            expected += [{at.element: list(at.coord/AU_to_A)}
                         for at in residue.get_atoms()]
        found = System()
        for frag in self.unmatched_fragments:
            info['found'][frag] = {'centroid':
                                   list(self[frag].centroid*AU_to_A),
                                   'nat': len(self[frag]),
                                   'fragment': self[frag]}
            found[frag] = self[frag]
        if len(expected) > 0:
            info['lookup'] = found.compute_matching(atlist=expected)
        return info

    def charge_at_pH(self, ph, closed_shell=True):
        """
        Identify the charge that the system should have for a given pH.
        Employ the
        `~py:meth:Bio.Bio.SeqUtils.ProtParam.ProteinAnalysis.charge_at_ph`
        method of BioPython

        Args:
            ph (float): the value of the ph
            closed_shell (bool): defines if the charge should be rounded
                to the nearest even integer considering the total number
                of electrons of the system.
        Returns:
            float : the value of the charge
        """
        from Bio.SeqUtils.ProtParam import ProteinAnalysis as PA
        protein = PA(str(sum(self.chains).sequence))
        charge = protein.charge_at_pH(ph)
        if closed_shell:
            nelec = float(sum(f.nel for f in self.values()))
            total_elec = round(nelec - charge)
            if int(total_elec/2)*2 != int(total_elec) and closed_shell:
                return int(charge+0.5)
        else:
            return round(charge) if closed_shell else charge

    def override_fragment_association(self, fragment_names, new_names=None):
        """
        Force the association of the fragment provided by the fragment_names
        list into the data represented by the new_names.
        New names should be provided as 'Ch-RES:num' where ``Ch`` is the chain,
        ``RES`` is the aminoacid residue at the position ``num``.

        Args:
           fragments_names(list): list of the fragments to be reassociated.
               Should be present in the unmatched_fragment list.
           new_names(list): list of the new names to be produced,
               in fragment_names element order.
               If absent, the fragment names are employed.
        """
        previously_unknown_fragments = {residue_name(res): ires
                                        for ires, (frag, res) in enumerate(
                                        zip(self.frag_to_residues,
                                            self.residues))
                                        if frag is None}
        if new_names is None:
            new_names = fragment_names
        for oldfrag, newfrag in zip(fragment_names, new_names):
            if oldfrag not in self.unmatched_fragments:
                raise ValueError(
                    'The fragment "' + oldfrag +
                    '" is not present in the list of unmatched fragments')
            iold = self.fragment_names.index(oldfrag)
            if self.fragment_letters[iold] != 'X':
                raise ValueError(
                    'The fragment "' + oldfrag +
                    '" has already a letter associated with it')
            if newfrag not in previously_unknown_fragments:
                raise ValueError(
                     'The fragment "' + newfrag +
                     '" cannot be associated to a valid sequence fragment')
            ires = previously_unknown_fragments.pop(newfrag)
            self.frag_to_residues[ires] = oldfrag
            for ic, chain in enumerate(self.chains_to_residues):
                if ires in chain:
                    iseq = chain.index(ires)
                    self.sequences_to_fragments[ic][iseq] = oldfrag
                    break
            self.fragment_letters[iold] = residue_letter(
                    ires, [seq.sequence for seq in self.chains],
                    self.chains_to_residues)
            self.unmatched_fragments.remove(oldfrag)
        if hasattr(self, '_fragment_ids'):
            delattr(self, '_fragment_ids')


def graph_bond(fraglist, threshold, pairwise_bo, restrict_to=None):
    """
    Defines the connectivity matrix of the fragments

    Args:
       fraglist (list): the list of the fragment names
       threshold (float) : the threshold of the cumulative fragment bond order
            that is applied to define the connectivity
       pairwise_bo (dict): list of the fragment BO connections
       restrict_to (list): list of fragments to which the graph has to be
            restricted

    Returns:
       (matrix) : connectivity matrix of the graph
    """
    from numpy import zeros
    mat = zeros((len(fraglist), len(fraglist)))
    lookup = {y: x for x, y in enumerate(fraglist)}

    for i, fragid1 in enumerate(fraglist):
        if restrict_to is not None and fragid1 not in restrict_to:
            continue
        bo = pairwise_bo[fragid1]
        sorted_bo = sorted(bo, key=bo.get, reverse=True)

        remainder = sum(bo.values())
        for key in sorted_bo:
            mat[i, lookup[key]] = 1
            remainder -= bo[key]
            if remainder < threshold:
                break
    return mat


def get_BO_network(connectivity_matrix):
    """
    Defines the connectivity matrix of the fragments
    Args:
       connectivity_matrix (matrix-like): matrix of the connnectivities of
           the system
       pairwise_bo (dict): list of the fragment BO connections

    Returns:
        tuple: graph,target_nodes,target_edges.
             Graph to be plotted and the list of target nodes and edges
             identified by the restriction if present.
    """
    import networkx as NX
    nw = NX.from_numpy_matrix(connectivity_matrix)
    target_nodes = []
    target_edges = []
    for cc in NX.connected_components(nw):
        subgraph = nw.subgraph(cc)
        if len(cc) > 1:
            target_nodes += list(subgraph.nodes)
            target_edges += list(subgraph.edges)
    return nw, target_nodes, target_edges


def fragment_letters(fragnames, sequences, residue_list,
                     sequences_to_residues):
    """
    Give a list of the fragments in terms of their letter

    Args:
        fragnames (list): list of the names of the fragments
        residue_list (list): The fragments of the `system` in the order
              of the residues of the structure
        sequences_to residues (list): strings of the protein sequence
              in the FASTA sequence order
        sequences(list): Sequences of the system
    Returns:
        list: the labels of the fragments as letters. the label is set to 'X'
           if the letter is not recognized
    """
    letters = ['X'] * len(fragnames)
    for ifrag, frag in enumerate(fragnames):
        if frag not in residue_list:
            continue
        ires = residue_list.index(frag)
        letters[ifrag] = residue_letter(ires, sequences, sequences_to_residues)
    return letters


def residue_letter(ires, sequences, sequences_to_residues):
    """
    Find the letter associated with the residue

    Args:
        ires (int): id of the residue
        sequences(list): Sequences of the system
        sequences_to residues (list): strings of the protein sequence
              in the FASTA sequence order

    Returns:
        str: letter of the residue
    """
    for iseq, seq in enumerate(sequences_to_residues):
        if ires in seq:
            ilett = seq.index(ires)
            return sequences[iseq][ilett]


def interaction_strengths(fragments, target_fragments, pairwise_bo):
    """
    Define the interaction strengths of the fragment list with the
    others with respect to a set of target fragments

    Args:
        fragments (list) : the name of the fragments of the
            system in pairwise_bo
        target_fragments (list) : the fragments constituting the target region
        pairwise_bo (dict): the bond order between fragments

    Returns:
        array: the  sum of the weigths of the interactions of the fragments
            in the target region
    """
    import numpy as np
    lookup = [fragments.index(frag) for frag in target_fragments]
    BOtot = np.zeros(len(fragments))
    for ipair in lookup:
        frag_i = fragments[ipair]
        BOtot += np.array([pairwise_bo[frag_i][frag] for frag in fragments])
    if len(lookup) > 0:
        BOtot[np.array(lookup)] = np.nan
    return BOtot


def find_relevant_fragments(sys, data, criteria, columns=None):
    """
    Identify the fragments of the system that fulfill the criteria
    provided by the function in argument

    Args:
        sys (BigDFT.System): the System to identify
        data (array, Dataframe, dict): data to take care of. If ana array,
            it is considered in order of fragment_names.
            If it is a dataframe or a dict, the name of fragments are the keys.
        criteria (func): function that can be applied to the data items.
            The function is applied to each of the data elements and the
            criteria is assumed to be satisfied
            if any of the element fulfill the criteria
        columns (list): the fragment to focus the search to.
            Useful in the case of a dictionary
    Returns:
        list: fragments that fulfill the criteria
    """
    from BigDFT.IO import reorder_fragments
    accepted = []
    for ikey, key in enumerate(sys.fragment_names):
        if isinstance(data, dict):
            if key in data:
                if columns is None:
                    columns = data[key].keys()
                datarr = [data[key][val] for val in columns]
            else:
                datarr = []
        else:  # this assumes that the value is a scalar
            datarr = [data[ikey]]
        condition = any(criteria(d) for d in datarr)
        if condition:
            accepted.append(key)
    return reorder_fragments(accepted)


def _get_system(system=None, archive=None, filename=None,
                logfile=None, deepcopy=False, **kwargs):
    from copy import deepcopy as dc
    if system is not None:
        return dc(system) if deepcopy else system
    if archive is not None:
        sys = load(archive,
                   serialization_version=kwargs.get('serialization_version'),
                   options=kwargs.get('options', {}))
        return sys
    if filename is not None:
        sys = BioSystem(filename, **kwargs.get('options', {}))
    if logfile is not None:
        sys.set_qm_run(logfile)
    return sys


class BioSystemPopulation(BioSystem):
    """
    An ensemble of BioSystem. Useful to provide and analyze averaged quantities

    Args:
       systems(dict, list): BioSystems defining the populations.
           They can be provided as a dict of pdb files plus logfiles,
           or as a dictionary of systems,
           or as a dictionary of archives for serialization.
           The dictionary is composed as follows:
           'label': {'archive': <archive_filename>,
                     'system': <BioSystem instance>,
                     'filename': <pdbfile>,
                     'logfile': <logfile_path>,
                     'weight': weigth of the sample in the population,
                     'mapping': provides the expression of the fragment of
                         the representative in term of the fragments of the
                         item of the population}
            alternatively, a list can be employed, in which case the filling of
            the population is performed sequentially
       representative (str, int): label of the system representative of
           the population
       exclude_representative (bool): if True, the data of the population will
           be evaluated without considering the representative
       fragment_values (list): list of the internal fragment values that can
           be found inside each of the system. Can be left as an empty list
           to delegate  all the evaluation to the populations.
       to_evaluate(dict) : dictionary of functions to be employed to
           evaluate at the creation of a population.
           They must provide a quantity that is related to a system.
       chessboard_cutoffs (list): list of the cutoffs that will have
           to be employed for the fragmentation.
           Use an empty list to deactivate.
       initial_view (dict): view of the superunits fragmentation that will be
           employed to the representative in the lowest cutoff
       **kwargs: arguments to be passed at the loading of the system
    """
    def __new__(cls, systems, representative, **kwargs):
        from futile.Utils import merge_two_dicts
        sys = _get_system(deepcopy=True,
                          **merge_two_dicts(systems[representative], kwargs))
        sys.__class__ = BioSystemPopulation
        return sys

    def __init__(self, systems, representative,
                 fragment_values=['charge', 'dipole', 'purity'],
                 to_evaluate=None,
                 exclude_representative=False,
                 chessboard_cutoffs=[0.05, 0.045, 0.04, 0.035, 0.03, 0.025],
                 initial_view=None,
                 **kwargs):
        from BigDFT.Stats import Population
        self.populations_items = fragment_values
        self.dataframe_populations = ['bond_orders', 'interactions',
                                      'electrostatic_interactions']
        self.populations = {key: Population(labels=self.fragment_names)
                            for key in self.populations_items +
                            self.dataframe_populations}
        self.extra_populations = {}
        self.representative = representative
        self.exclude_representative = exclude_representative
        self._extra_population('system_dfs', None)
        if to_evaluate is not None:
            for label, func in to_evaluate.items():
                self._extra_population(label, func)
        self.systems = systems
        if len(chessboard_cutoffs) > 0:
            self.chessboard_dict = super(
                BioSystemPopulation, self).chessboards(
                cutoffs=chessboard_cutoffs, initial_view=initial_view)
        else:
            self.chessboard_dict = None
        self._fill_populations(**kwargs)

    def _fill_populations(self, **kwargs):
        from futile.Utils import fill_dictionary_in_parallel, kw_pop
        new_kwargs, nthreads = kw_pop('nthreads', 1, **kwargs)
        new_kwargs['biopop'] = self
        new_kwargs['cbs'] = self.chessboard_dict
        if nthreads == 1:
            dict_sys = {}
            if isinstance(self.systems, dict):
                keys = [k for k in self.systems.keys()
                        if not self.exclude_representative or
                        k != self.representative]
            else:
                keys = range(len(self.systems))
                if self.exclude_representative:
                    keys.remove(self.representative)
            for label in keys:
                dict_sys[label] = _data_of_one_system(label, **new_kwargs)
        else:
            dict_sys = fill_dictionary_in_parallel(nthreads,
                                                   self.systems.keys(),
                                                   _data_of_one_system,
                                                   **new_kwargs)
        for results in dict_sys.values():
            for pop_label, pop in self.populations.items():
                pop.append(**results[pop_label])
        self._assign_chessboard()

    def _extra_population(self, label, func):
        from BigDFT.Stats import Population
        self.populations[label] = Population()
        self.populations_items.append(label)
        self.extra_populations[label] = func

    def _assign_chessboard(self):
        if self.chessboard_dict is not None:
            self._cached_refragment = {str(c): clean_view(cb[-1], self)
                                       for c, cb in
                                       self.chessboard_dict.items()}

    def serialize(self, archive, chessboard_dict=None):
        """
        Dump the entire population into an archive, containing
        the representative archive, if it is provided as such
        (otherwise the serialization of the system)
        as well as the numpy files of all the population data.

        Args:
             archive(str): archive to dump the population to
             chessboard_dict (dict): optional dictionary of the fragmentation
                 of the population that has been previously found with
                 a chessboard algorithm. If absent, the internal chessboard is
                 employed
        """
        from BigDFT.Stats import dump_populations
        from futile.Utils import serialize_objects, create_tarball
        if 'archive' in self.systems[self.representative]:
            files = [self.systems[self.representative]['archive']]
        elif 'filename' in self.systems[self.representative]:
            files = [self.systems[self.representative]['filename']]
        else:
            self._cached_refragment = {str(c): cb[-1]
                                       for c, cb in chessboard_dict.items()}
            serialization = BioSystemSerialization(self, version='1.1')
            archive = 'representative.tar.bz2'
            serialization.dump(archive)
            files = [archive]
        files += list(dump_populations(self.populations))
        if chessboard_dict is None:
            chessboard_dict = self.chessboard_dict
        if chessboard_dict is not None and len(chessboard_dict) > 0:
            chb_obj = serialize_objects({'chessboard.json': chessboard_dict})
        else:
            chb_obj = {}
        create_tarball(archive, set(files), chb_obj)

    @classmethod
    def load(cls, archive, **kwargs):
        """
        Create a BioSystemPopulation instance from a serialized archive.

        Args:
            archive (str): the path of the archive
            **kwargs: other arguments that have to be passed to the
             BioSystemPopulation instantiation

        Returns:
            BioSystemPopulation: a biosystem population instance
        """
        from futile.Utils import unpack_tarball
        from shutil import rmtree
        from os.path import join
        from BigDFT.Stats import Population
        import json

        tmpdir, files = unpack_tarball(archive)

        for f in files:
            if '.tar.bz2' in f:
                # we found the system
                sys_dict = [{'archive': join(tmpdir, f)}]
                break
            if '.pdb' in f:
                # we found the filename
                sys_dict = [{'filename': join(tmpdir, f)}]
        loaded = cls(systems=sys_dict, representative=0,
                     fragment_values=[], chessboard_cutoffs=[], **kwargs)
        for f in files:
            if '.json' in f:
                ff = open(join(tmpdir, f), 'r')
                loaded.chessboard_dict = json.load(ff)
                loaded._assign_chessboard()
                break

        pop_data = {f.replace('.npy', ''): Population.load(join(tmpdir, f))
                    for f in files if '.npy' in f}
        loaded.populations = pop_data

        rmtree(tmpdir)
        return loaded

    def _present_as_population(self, method):
        return hasattr(self, 'populations') and method in self.populations \
            and len(self.populations[method].datas) > 0

    @property
    def purities(self):
        if self._present_as_population('purity'):
            pop = self.populations['purity']
            mean = {f: -pop.mean[i] for i, f in enumerate(pop.feature_labels)}
            return mean
        else:
            return super(BioSystemPopulation, self).purities

    @property
    def bond_orders(self):
        if self._present_as_population('bond_orders'):
            pop = self.populations['bond_orders']
            return pop.mean.to_dict()
        else:
            return super(BioSystemPopulation, self).bond_orders

    @property
    def interactions(self):
        if self._present_as_population('interactions'):
            pop = self.populations['interactions']
            return pop.mean.to_dict()
        else:
            return super(BioSystemPopulation, self).interactions

    def fragment_values(self, criteria, view=None):
        if criteria is None:
            return None
        field_vals = self.populations[criteria].mean
        errors = self.populations[criteria].std
        if view is not None:
            field_vals = _aggregate_values(self.fragment_names,
                                           field_vals, view)
        return {'field_vals': field_vals, 'errors': errors}

    def colordict(self, field_vals=None, **kwargs):
        if field_vals is None:
            field_vals = self.fragment_values(criteria=kwargs.get('color_by'))
            if field_vals is not None:
                field_vals = field_vals['field_vals']
        elif isinstance(field_vals, dict) and 'field_vals' in field_vals:
            field_vals = field_vals['field_vals']
        return super(BioSystemPopulation, self).colordict(
            field_vals=field_vals, **kwargs)

    def fragment_scatterplot(self, field_vals=None, errors=None, **kwargs):
        colors = self.colordict(field_vals=field_vals, **kwargs)
        if errors is None and field_vals is None:
            errors = self.fragment_values(
                                    criteria=kwargs.get('color_by'))['errors']
        if field_vals is None:
            field_vals = self.fragment_values(
                                 criteria=kwargs.get('color_by'))['field_vals']
        return super(BioSystemPopulation, self).fragment_scatterplot(
            field_vals=field_vals, errors=errors, colordict=colors, **kwargs)


def _data_of_one_system(label, **kwargs):
    from futile.Utils import merge_two_dicts, write
    biopop = kwargs['biopop']
    args = biopop.systems[label]
    wg = args.get('weight', 1.0)
    verbose = kwargs.get('verbose', True)
    if verbose:
        write(str(label) + ':  # weight='+str(wg))
    sys = _get_system(**merge_two_dicts(args, kwargs))
    results = {}
    # fragment the population
    cbs = kwargs.get('cbs')
    cutoffs = list(cbs.keys()) if cbs is not None else []
    if len(cutoffs) > 0:
        loosest_coff = sorted(cutoffs, reverse=True)[0]
        previous_view = cbs[loosest_coff][-1]
        new_view = clean_view(previous_view, sys)
        sys_cbs = sys.chessboards(cutoffs, initial_view=new_view)
    else:
        sys_cbs = []
    for cutoff in sys_cbs:
        view = cbs[cutoff][-1]
        review = sys_cbs[cutoff][0]
        # review = update_fragmentation(cutoff, view, sys)
        cbs[cutoff].append(review)
        if len(view) != len(review):
            write('The number of fragments has been modified', len(view),
                  len(review))
    for pop_label in biopop.populations:
        of = pop_label
        if verbose:
            write('  - ' + str(pop_label))
        if of in biopop.extra_populations and of != 'system_dfs':
            array = biopop.extra_populations[of](sys)
        elif of in biopop.dataframe_populations:
            array = getattr(sys, of, 0.0)
        else:
            array = sys.fragment_values(of)
        if pop_label == 'system_dfs':
            order = [frag for frag in biopop.fragment_names if frag in sys]
            order += [frag for frag in sys if frag not in order]
            data = sys.to_dataframe(order=order)
        else:
            data = reorder(array, sys, biopop, mapping=kwargs.get('mapping'))
        results[pop_label] = dict(data=data, weight=wg, label=label)
    return results


def update_fragmentation(cutoff, view_old, sys):
    """
    Update the fragmentation views starting from the last element of
    a list of views

    Args:
         cutoff (float): the cutoff to employ for the fragmentation
         view_old (dict): previous view of the system
         sys (BioSystem): the system employed to update the frament list
    Returns:
         dict: new view of the system
    """
    view = clean_view(view_old, sys)
    # erase any previously existing refragmentation
    sys._cached_refragment = {}
    return sys.refragment(cutoff, view=view)


def reorder(array, of, with_respect_to, switch_chains=None, mapping=None):
    """
    Reorder an array of data of systems in such a way that it can be applied
    to characterize another one.

    Args:
        array (array-like, dict): data in the order of `of.fragment_names`.
            in case of a dictionary, it is assumed that it contains the data
            coming from interactions or bond orders.
        of (BioSystem): the system associated to the original data
        with_respect_to (BioSystem): the system to which reorder the data
        switch_chains (dict): mapping of the chains of the old system with
           respect to the new.
        mapping (dict): dictionary of the view that decomposes the fragments of
           ``with_respect_to`` in terms of the fragment of ``of`` in case of
           multiple elements in the values, an aggregation is considered.
           If mapping is present, there should not be undefined fragments
           in their values

    Returns:
        numpy.array: data in the new order. It contains `numpy.nan` for the
            fragments which are not present in the new system.
    """
    from numpy import ones, nan
    from BigDFT.PostProcessing import superunits_quadratic_quantities
    if isinstance(array, dict):
        if mapping is not None:
            newdata = superunits_quadratic_quantities(array, mapping)
        else:
            newdata = array
            # newdata = {}
            # for frag in with_respect_to.fragment_names:
            #     newdata[frag] = {f: array[frag].get(f, 0)
            #                      for f in with_respect_to.fragment_names}
        return newdata
    data = ones(len(with_respect_to.fragment_names))*nan
    if mapping is not None:
        newarray = _aggregate_values(of.fragment_names, array, view=mapping)
    else:
        newarray = array
    for ifrag, frag_orig in enumerate(with_respect_to.fragment_names):
        if switch_chains is None:
            frag = frag_orig
        else:
            frag = None
            for ch_o, ch_n in switch_chains.items():
                if ch_o not in frag_orig:
                    continue
                frag = frag_orig.replace(ch_o+'-', ch_n+'-')
                break
            if frag is None:
                frag = frag_orig
        if frag not in of.fragment_names:
            continue
        jfrag = of.fragment_names.index(frag)
        data[ifrag] = newarray[jfrag]
    return data


def clean_view(view_orig, superunits):
    """
    Define a view of the system that comes from a view of another one.
    Superunits are cleaned in such a way that no inconsistencies are present
    in the new view.

    Args:
       view_orig (dict): the original view we would like to clean
       superunits (list): list of the new system's superunits
    Returns:
       dict: the new view of the system, where combined superunits are
           joined with the '+' symbol
    """
    if view_orig is None:
        return None
    view = view_orig.copy()
    allvals = []
    for k, val_orig in view_orig.items():
        val = [v for v in val_orig]
        for v in val_orig:
            if v not in superunits:
                previous_k = '+'.join(val)
                if k in view:
                    view.pop(k)
                else:
                    view.pop(previous_k)
                val.remove(v)
                rek = '+'.join(val)
                view[rek] = val
            else:
                allvals.append(v)
    for k in list(view.keys()):
        if len(view[k]) == 0:
            view.pop(k)
    for k in superunits:
        if k not in allvals:
            view[k] = [k]
    return view


def rename_residue(res):
    """
    Rename the residue of the system such that alphabetical order
    corresponds to sequence order

    Args:
        res (str): residue of the system, in form "chain-resname:pos"
    Returns:
        str: residue in the form "chain-pos:letter"
    Warning:
        Only put letters to residues which have letter associated
    """
    from Bio.PDB.Polypeptide import three_to_one
    ch, namepos = res.split('-')
    name, pos = namepos.split(':')
    try:
        letter = three_to_one(name)
    except Exception:
        letter = name
    rename = ch+'-'+str(pos).zfill(3)+':'+letter
    return rename


def rename_labels(labels, as_dict=False, from_dict=False,
                  mappable=rename_residue):
    """
    Rename the names of the residues of a BioSystem in order
    to be compatible with other naming schemes.

    Args:
        labels (iterable): contains the list of residue to be renamed
        as_dict (bool): if True, returns a dictionary as a relabel
        from_dict (bool): if True assumes that the remapping has to be
           performed also on the labels values.
        mappable (func): the function that will be applied for renaming
    Returns:
        dict, list: dictionary or list of the renamed labels, depending on
            the arguments.
    """
    relabel = [] if not from_dict else {}
    for label in labels:
        rename = '+'.join([mappable(res) for res in label.split('+')])
        if from_dict:
            relabel[rename] = map(mappable, labels[label])
        else:
            relabel.append(rename)
    return {k: v for k, v in zip(labels, relabel)} if as_dict else relabel



"""
This module will generate a set of python unit tests that will automatically
run each of the `_example` subroutines in the PyBigDFT package.

python -m unittest discover -s $PYTHONPATH/BigDFT/ -p "check_examples.py"
"""
from pkgutil import walk_packages
from futile.Utils import write as safe_print
from future.utils import with_metaclass
from unittest import TestCase
import BigDFT
from contextlib import redirect_stdout


class MetaTest(type):
    """
    A meta test generator that creates test cases from the test spec.
    """
    def __new__(mcs, name, bases, dict):
        # Generate a closure based on a given module.
        def gen_test(name, module):
            def test(self):
                with open(name + ".txt", "w") as ofile:
                    with redirect_stdout(ofile):
                        safe_print("Running Test:", name)
                        module._example()
            return test

        # This dictionary will be a map from test names to test methods.
        dict = {}

        # Walk over all packages
        for _, modname, _ in walk_packages(path=BigDFT.__path__,
                                           prefix=BigDFT.__name__+'.',
                                           onerror=lambda x: None):
            try:
                # Import the module we found
                module = __import__(modname, fromlist="dummy")

                # See if it has an example routine.
                if hasattr(module, "_example"):
                    name = modname.replace(".", "_").replace("/", "")
                    if "Interop" in name:  # for now we disable interop
                        continue
                    dict["test_" + name] = gen_test(name, module)
            except ModuleNotFoundError:
                continue

        return type.__new__(mcs, name, bases, dict)


class ExampleTest(with_metaclass(MetaTest, TestCase)):
    """
    The actual test class. The MetaTest provides the test cases, and inheriting
    from TestCase lets it be automatically be run using ``unittest``.
    """
    pass



"""
This module contains some wrappers for using OpenBabel to perform
various operations on BigDFT molecules.

https://open-babel.readthedocs.io/en/latest/UseTheLibrary/Python.html
"""

# openbabel's forcefields can have different energy units.
_energy_conversion = {"kJ/mol": 0.00038, "kcal/mol": 0.0016}


def convert_system_to_babel(sys):
    """
    Convert a BigDFT system to an open babel molecule.

    Args:
      frag (BigDFT.Systems.System): the system to convert.

    Returns:
      (openbabel.OBMol): an open babel type molecule.
    """
    from BigDFT.IO import write_pdb
    from openbabel.openbabel import OBMol, OBConversion
    # py2 workaround
    from sys import version_info
    if version_info[0] < 3:
        from io import BytesIO as StringIO
    else:
        try:
            from io import StringIO
        except ImportError:
            from StringIO import StringIO

    # We convert by way of pdb file.
    conv = OBConversion()
    conv.SetInFormat("pdb")

    sval = StringIO()
    write_pdb(sys, sval)

    mol = OBMol()
    conv.ReadString(mol, sval.getvalue())

    return mol


def convert_babel_to_system(mol):
    """
    Convert a BigDFT fragment to an open babel molecule.

    Args:
      frag (BigDFT.Fragments.Fragment): the fragment to convert.

    Returns:
      (openbabel.OBMol): an open babel type molecule.
    """
    from BigDFT.IO import read_mol2
    from openbabel.openbabel import OBConversion
    # py2 workaround
    from sys import version_info
    if version_info[0] < 3:
        from io import BytesIO as StringIO
    else:
        try:
            from io import StringIO
        except ImportError:
            from StringIO import StringIO

    conv = OBConversion()
    conv.SetOutFormat("mol2")

    sval = StringIO(conv.WriteString(mol))
    return read_mol2(sval)


def compute_smiles(sys):
    """
    Computes the SMILES representation of a given system.

    Args:
      sys (BigDFT.System.Systems): the system to compute the
        representation of.

    Return:
      (str): the smiles representation of this molecule.
    """
    from openbabel.openbabel import OBConversion

    conv = OBConversion()
    mol = convert_system_to_babel(sys)
    conv.SetOutFormat("SMI")

    retstr = conv.WriteString(mol)
    retstr = retstr.replace("\n", "")
    retstr = retstr.replace("\t", "")

    return retstr


def compute_fingerprint(sys, fps="fp2"):
    """
    Computes the fingerprint for a particular fragment.

    Args:
      sys (BigDFT.Systems.System): the fragment to compute the
        representation of.
      fps (str): the type of finger print to compute.

    Return:
      (openbabel.OBFingerprint): a fingerprint for this fragment.
    """
    from pybel import Molecule

    mol = convert_system_to_babel(sys)
    pybelmol = Molecule(mol)

    return pybelmol.calcfp(fps)


def system_energy(sys, forcefield="MMFF94", verbose=False):
    """
    Compute the energy of a system using an openbabel forcefield.

    Args:
      frag (BigDFT.Systems.System): the system to compute.
      forcefield (str): the type of forcefield to use.
      verbose (bool): whether to have openbabel run in verbose mode.

    Returns:
      (float): the energy value computed in Hartree.
    """
    from openbabel.openbabel import OBForceField, OBFF_LOGLVL_LOW

    # Setup the forcefield
    ff = OBForceField.FindForceField(forcefield)
    mol = convert_system_to_babel(sys)
    ff.Setup(mol)

    if verbose:
        ff.SetLogToStdOut()
        ff.SetLogLevel(OBFF_LOGLVL_LOW)

    # Call the energy routine.
    return ff.Energy() * _energy_conversion[ff.GetUnit()]


def optimize_system(sys, forcefield="MMFF94", method="SteepestDescent",
                    steps=1000, econv=1e-6, verbose=False):
    """
    Optimize the geometry of a given fragment.

    Args:
      frag (BigDFT.Systems.System): the fragment to optimize.
      forcefield (str): the type of forcefield to use.
      verbose (bool): if True, the openbabel output will be printed.

    Returns:
      (BigDFT.Fragments.Fragment): a new fragment with the optimized positions.
    """
    from openbabel.openbabel import OBForceField, OBFF_LOGLVL_LOW
    from copy import deepcopy

    # Setup the forcefield
    ff = OBForceField.FindForceField(forcefield)
    mol = convert_system_to_babel(sys)
    ff.Setup(mol)
    if verbose:
        ff.SetLogToStdOut()
        ff.SetLogLevel(OBFF_LOGLVL_LOW)

    # Call the optimization routine.
    if method == "SteepestDescent":
        ff.SteepestDescent(steps, econv)
    elif method == "ConjugateGradients":
        ff.ConjugateGradients(steps, econv)
    else:
        raise ValueError("Invalid minimization method.")

    # Extract out the the positions.
    ff.GetCoordinates(mol)
    newsys = convert_babel_to_system(mol)

    return newsys


def molecular_dynamics(sys, steps, temperature, forcefield="MMFF94",
                       timestep=0.001, verbose=False):
    """
    Run molecular dynamics on a given fragment..

    Args:
      sys (BigDFT.Systemtems.System): the system to run.
      steps (int): the number of MD steps to take.
      temperature (float): temperature in K.
      timestep (float): time step in picoseconds.
      forcefield (str): the type of forcefield to use.
      constraints (list): for each atom, list whether it if frozen or not.
      verbose (bool): if True, the openbabel output will be printed.

    Returns:
      (BigDFT.Systems.System): a new system with the optimized positions.
    """
    from openbabel.openbabel import OBForceField, OBFF_LOGLVL_LOW

    # Setup the calculation
    ff = OBForceField.FindForceField(forcefield)
    mol = convert_system_to_babel(sys)
    ff.Setup(mol)
    if verbose:
        ff.SetLogToStdOut()
        ff.SetLogLevel(OBFF_LOGLVL_LOW)

    ff.MolecularDynamicsTakeNSteps(steps, temperature, timestep)

    ff.GetCoordinates(mol)
    newsys = convert_babel_to_system(mol)

    return newsys


def _setup_constraints(forcefield, constraints):
    """
    This helper routine takes a list of constraints and updates
    as forcefield with those values.
    """
    constr = forcefield.GetConstraints()
    for i in range(0, len(constraints)):
        if constraints[i] is None:
            continue
        elif constraints[i] == "fxyz" or constraints[i] == "f":
            constr.AddAtomConstraint(i+1)
        elif constraints[i] == "fx":
            constr.AddAtomXConstraint(i+1)
        elif constraints[i] == "fy":
            constr.AddAtomYConstraint(i+1)
        elif constraints[i] == "fz":
            constr.AddAtomZConstraint(i+1)
    forcefield.SetConstraints(constr)


def _example():
    """Example of using OpenBabel interoperability"""
    from BigDFT.Systems import System
    from BigDFT.Fragments import Fragment
    from BigDFT.IO import XYZReader
    from os.path import join

    # Read in a system.
    sys = System()
    sys["FRA:1"] = Fragment()
    with XYZReader("CH4") as ifile:
        for at in ifile:
            sys["FRA:1"] += Fragment([at])

    # We can compute the smiles representation.
    print(compute_smiles(sys))

    # The energy.
    print(system_energy(sys, forcefield="UFF"))

    # Optimize the geometry.
    sys2 = optimize_system(sys, forcefield="UFF")
    print(system_energy(sys2, forcefield="UFF"))


if __name__ == "__main__":
    _example()



"""
A module for handling unit cells for periodic calculations.
"""


class UnitCell:
    """
    Defines a wrapper for unit cells.

    Args:
        cell (list): a list of unit cell vectors. This can either be a 3x1 list
          or a 3x3 list. Currently, only orthorhombic cells are supported.
          units (str): the units of the cell parameters. If cell is set to
          None, the free boundary condition is enforced.
        units (str): the unit of length.
    """
    def __init__(self, cell=None, units="bohr"):
        from BigDFT.Atoms import AU_to_A, IsAngstroem

        # Early exit for free boundary condition
        if cell is None:
            self.cell = [[float("inf"), 0, 0],
                         [0, float("inf"), 0],
                         [0, 0, float("inf")]]
            return

        # Copy cell based on the format
        if isinstance(cell[0], list):
            self.cell = cell
        else:
            self.cell = [[cell[0], 0, 0], [0, cell[1], 0], [0, 0, cell[2]]]

        # Check that the unit cell is valid
        if self[2, 2] == float("inf") and not (self[1, 1] == float("inf") and
           self[0, 0] == float("inf")):
            raise ValueError("Boundary condition: ",
                             "If z is infinity, x and y must also be.")
        elif self[0, 0] == float("inf") and self[1, 1] != float("inf"):
            raise ValueError("Boundary condition: ",
                             "If x is infinity, y must also be.")
        elif self[0, 1] != 0 or self[0, 2] != 0 or \
                self[1, 0] != 0 or self[1, 2] != 0 or \
                self[2, 0] != 0 or self[2, 1] != 0:
            raise ValueError("Non-orthorhombic cells not supported")

        for i in range(3):
            for j in range(3):
                if self[i, j] < 0:
                    raise ValueError("Unit cell must be non-negative values.")

        # Store internally in bohr
        if IsAngstroem(units):
            for i in range(3):
                for j in range(3):
                    self.cell[i][j] /= AU_to_A

    def __getitem__(self, idx):
        return self.cell[idx[0]][idx[1]]

    def get_boundary_condition(self, units="bohr"):
        """
        Get a string description of the boundary condition (i.e. free,
        surface, wire, periodic)
        """
        from BigDFT.Atoms import AU_to_A, IsAngstroem

        # Match units
        if IsAngstroem(units):
            conv = AU_to_A
        else:
            conv = 1

        if all([self[i, i] == float("inf") for i in range(3)]):
            return "free"
        elif self[0, 0] == float("inf") and self[1, 1] == float("inf"):
            return "wire 0.0 0.0 " + str(self[2, 2]*conv)
        elif self[1, 1] == float("inf"):
            return "surface " + str(self[0, 0]*conv) + " 0.0 " + \
                   str(self[2, 2]*conv)
        else:
            return "periodic " + " ".join([str(conv*self[i, i])
                                           for i in range(3)])

    def get_posinp(self, units="bohr"):
        """
        Create the dictionary representation of the cell that is passed to
        BigDFT.
        """
        from BigDFT.Atoms import AU_to_A, IsAngstroem
        if IsAngstroem(units):
            return [self[i, i] * AU_to_A for i in range(3)]
        else:
            return [self[i, i] for i in range(3)]

    def minimum_image(self, pos, units="bohr"):
        """
        Given a vector of three positions, this wraps those positions inside
        the cell using the minimum image convention.
        """
        from BigDFT.Atoms import AU_to_A, IsAngstroem

        # Match units
        if IsAngstroem(units):
            conversion = 1/AU_to_A
        else:
            conversion = 1
        bohrpos = [x*conversion for x in pos]

        # Adjust position
        for i in range(3):
            if self[i, i] == float("inf"):
                continue
            while(bohrpos[i] > self[i, i]):
                bohrpos[i] -= self[i, i]
            while(bohrpos[i] < 0):
                bohrpos[i] += self[i, i]

        # Convert back
        return [x/conversion for x in bohrpos]


def _example():
    # Create a basic unit cell
    cell = UnitCell([10, 8, 4], units="angstroem")

    # Print out the posinp representation
    print(cell.get_posinp())
    print(cell.get_posinp(units="angstroem"))

    # Right now we enforce the orthorhombic condition
    try:
        cell = UnitCell([[10, 0, 0], [0, 8, 0], [0, 4, 10]])
    except ValueError as e:
        print(e)
    cell = UnitCell([[10, 0, 0], [0, 8, 0], [0, 0, 4]])
    print(cell.get_boundary_condition("angstroem"))

    # Wire boundary condition
    wire = UnitCell([float("inf"), float("inf"), 4])
    print(wire.get_posinp())
    print(wire.get_boundary_condition())

    # Surface boundary condition
    surface = UnitCell([10, float("inf"), 4])
    print(surface.get_posinp())
    print(surface.get_boundary_condition())

    # Wrap positions to the minimum image convention.
    pos = [-5, -2, -3]
    print(cell.minimum_image(pos))
    print(wire.minimum_image(pos))
    print(surface.minimum_image(pos))

    pos = [15, 12, 13]
    print(cell.minimum_image(pos))
    print(wire.minimum_image(pos))
    print(surface.minimum_image(pos))


if __name__ == "__main__":
    _example()



"""
This module contains operations for the treatment of quantities related
to particles in space.
"""


def distance_matrix(X):
    """
    Extract the distance matrix of a array of positions with euclidean
    metric.

    Args:
        X (array): array of the positions of the system

    Returns:
        matrix-like: matrix of distances in squareform

    Warning:
        The distance provided here is meaningful only in isolated boundary
        conditions
    """
    from scipy.spatial import distance as d
    compactform = d.pdist(X)
    return d.squareform(compactform)


def inverse_distance_matrix(Rab):
    """
    Calculate the inverse distance matrix, with zeros on the diagonal

    Args:
         Rab (matrix-like): matrix of the atomic distances in squareform
    Returns:
         matrix-like: inverse distance matrix, with diagonal values set to 0
    """
    from numpy import seterr, fill_diagonal
    err = seterr(divide='ignore')
    rabm1 = 1.0/Rab
    seterr(**err)
    fill_diagonal(rabm1, 0.0)
    return rabm1


def electrostatic_energies_matrix(X, Rm1, Q, P):
    """
    Extract the matrix of the electrostatic energies of atoms described
    By their positions, charges and dipoles

    Args:
         X (array): the array of the atom positions
         Q (array): charge monopoles of the atoms (total)
         P (array): array of the dipoles, in Px,Py,Pz convention
         Rm1 (matrix): inverse distance matrix of the positions

    Returns:
         matrix: matrix of the electrostatic energies per atom pair

    Warning:
         This energy is meaningful only for isolated boundary conditions
    """
    from numpy import array, diag
    E1 = Q.dot(Q.T)
    E1 *= Rm1
    if P is not None:
        E3 = P.dot(P.T)
        E3 *= Rm1**3

        PX = P.dot(X.T)
        D = array(diag(PX))
        PX = (D-PX.T).T
        PX *= Rm1**2
        PXQ = PX*Q.T[0]
        PX *= PX.T

        PXQ += PXQ.T
        # PXQ -= 3.0*PX
        PXQ *= Rm1

        E3 = E3 + 3.0*PX*Rm1 - PXQ
    else:
        E3 = 0.0
    return 0.5 * (E1 + E3)


def long_range_energy_matrix(X, Rm1, Z, Q, P):
    """
    Extract the matrix of the electrostatic energies of atoms described
    By their positions, charges and dipoles

    Args:
         X (array): the array of the atom positions
         Q (array): charge monopoles of the electrons
         Z (array): charge monopole of the ions
         P (array): array of the dipoles, in Px,Py,Pz convention
         Rm1 (matrix): inverse distance matrix of the positions

    Returns:
         matrix: matrix of the electrostatic energies per atom pair

    Warning:
         This energy is meaningful only for isolated boundary conditions
    """
    from numpy import array, diag
    QQ = Q + Z
    E1 = QQ.dot(Q.T)
    E1 += E1.T
    E1 *= Rm1
    if P is not None:
        E3 = P.dot(P.T)
        E3 *= Rm1**3

        PX = P.dot(X.T)
        D = array(diag(PX))
        PX = (D-PX.T).T
        PX *= Rm1**2
        PXQ = PX*QQ.T[0]
        PX *= PX.T

        PXQ += PXQ.T
        # PXQ -= 3.0*PX
        PXQ *= Rm1

        E3 = 2.0 * E3 + 3.0*PX*Rm1 - PXQ
    else:
        E3 = 0.0
    return -0.5 * (E1 + E3)


def electrostatic_energy_dict(E, slicing):
    """
    Dictionary of the electrostatic energies starting from a slicing
    defining the blocks of atom to be considered

    Args:
        E (matrix): electrostatic_energies_matrix of the atoms
        slicing (dict): slicing of the fragments, given for example
            by `py:meth:~BigDFT.Systems.System.dataframe_slicing`
    Returns:
        dict: dictionary of the electrostatic interaction per blocks
    """
    eldict = {}
    for frag1 in slicing:
        for frag2 in slicing:
            if frag2 in eldict and frag1 in eldict[frag2]:
                e = eldict[frag2][frag1]
            else:
                s1, e1 = slicing[frag1]
                s2, e2 = slicing[frag2]
                e = E[s1:e1, s2:e2].sum()
                # E[group1][:, group2].sum() useful for non-contiguous
            eldict.setdefault(frag1, {})[frag2] = e
    return eldict


class PointParticles():
    """
    An object storing the point particle information

    Args:
        X (array): array of atomic positions
        Q (array): charge monopoles of the electrons
        Z (array): charge monopole of the ions
        P (array): electrostatic dipoles
    """
    def __init__(self, X, Z=None, Q=None, P=None):
        self.X = X
        if Z is not None:
            self.Z = Z
        if Q is not None:
            self.Q = Q
        self.P = P

    @property
    def R(self):
        """
        Distance matrix between point PointParticles
        """
        if not hasattr(self, '_R'):
            self._R = distance_matrix(self.X)
        return self._R

    @property
    def Rm1(self):
        """
        Inverse Distance matrix
        """
        if not hasattr(self, '_Rm1'):
            self._Rm1 = inverse_distance_matrix(self.R)
        return self._Rm1

    @property
    def Eel(self):
        """
        Matrix of the pair atomic electrostatic energies
        """
        if not hasattr(self, '_Eel'):
            self._Eel = electrostatic_energies_matrix(self.X, self.Rm1,
                                                      self.Q + self.Z, self.P)
        return self._Eel

    @property
    def Elr(self):
        """
        Matrix of the long range energy
        """
        if not hasattr(self, '_Elr'):
            self._Elr = long_range_energy_matrix(self.X, self.Rm1,
                                                 self.Z, self.Q, self.P)
        return self._Elr

    def Eel_dict(self, slicing):
        """
        Dictionary of the electrostatic energies given a slicing defining
        the groups of atoms to gather
        """
        return electrostatic_energy_dict(self.Eel, slicing)

    def Elr_dict(self, slicing):
        """
        Dictionary of the electrostatic energies given a slicing defining
        the groups of atoms to gather
        """
        return electrostatic_energy_dict(self.Elr, slicing)
